import datetime
from datetime import date

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.optimize

def xnpv(rate, values, dates):
    '''Equivalent of Excel's XNPV function.

    >>> from datetime import date
    >>> dates = [date(2010, 12, 29), date(2012, 1, 25), date(2012, 3, 8)]
    >>> values = [-10000, 20, 10100]
    >>> xnpv(0.1, values, dates)
    -966.4345...
    '''
    if rate <= -1.0:
        return float('inf')
    d0 = dates[0]    # or min(dates)
    return sum([ vi / (1.0 + rate)**((di - d0).days / 365.0) for vi, di in zip(values, dates)])

def xirr(values, dates):
    '''Equivalent of Excel's XIRR function.

    >>> from datetime import date
    >>> dates = [date(2010, 12, 29), date(2012, 1, 25), date(2012, 3, 8)]
    >>> values = [-10000, 20, 10100]
    >>> xirr(values, dates)
    0.0100612...
    '''
    try:
        return scipy.optimize.brentq(lambda r: xnpv(r, values, dates), -0.999, 1e10)
    except RuntimeError:    # Failed to converge?
        return scipy.optimize.newton(lambda r: xnpv(r, values, dates),  x0=0.0, maxiter=40)


class AIPCalculator(object):
    def __init__(self, arr, start_date, end_date, stop_profit=None, strategy=None,
                 amt_bench=1000, mode='1w', mode_offset=0, purchase_rate=0.001, redeem_rate=0.001):
        self.arr = arr.dropna()
        self.start_date = max(start_date, self.arr.index[0])
        self.end_date = end_date or datetime.datetime.now().date()
        self.amt_bench = amt_bench
        self.mode = mode
        self.mode_offset = mode_offset
        self.yearly_multiplier = self.get_yearly_multiplier(mode)
        self.purchase_rate = purchase_rate
        self.redeem_rate = redeem_rate
        self.stop_profit = stop_profit
        # calc
        self.last_trade_day = None
        self.moving_average = None
        if strategy:
            if strategy.startswith('ma'):
                window_size = int(strategy[2:])
                self.moving_average = self.arr.rolling(window_size).mean().fillna(0)
                self.vol_10 = self.arr.pct_change(1).rolling(10).std().fillna(0)

    def get_yearly_multiplier(self, mode):
        m = None
        if mode == '1d':
            m = 365
        elif mode == '1w':
            m = 52
        elif mode == '2w':
            m = 26
        elif mode == '1m':
            m = 12
        else:
            raise Exception(f'unexpected trade mode: {mode}')
        return m

    # 判断是否当前要进行交易
    def is_to_trade(self, dt):
        if self.mode == '1d':
            return True
        elif self.mode == '1w':
            return dt.weekday() == self.mode_offset
        elif self.mode == '2w':
            return dt.weekday() == self.mode_offset and (
            self.last_trade_day is None or (dt - self.last_trade_day).days > 7)
        elif self.mode == '1m':
            return dt.day == self.mode_offset + 1
        else:
            raise Exception('unexpected trade mode: ' + self.mode)

    def convert_irr(self, irr):
        return pow(irr + 1, self.yearly_multiplier) - 1

    def convert_ret(self, ret):
        days = (self.end_date - self.start_date).days
        return pow(ret + 1, 365 / days) - 1

    def _mv_multiplier(self, val, ma, vol_10):
        if ma == 0:
            return 1
        diff = (val - ma) / ma
        if diff > 0:
            if diff <= 0.15:
                return 0.9
            elif diff <= 0.5:
                return 0.8
            elif diff <= 1:
                return 0.7
            else:
                return 0.6
        else:
            diff = abs(diff)
            if diff <= 0.05:
                return 0.6 if vol_10 > 0.05 else 1.6
            elif diff <= 0.1:
                return 0.7 if vol_10 > 0.05 else 1.7
            elif diff <= 0.2:
                return 0.8 if vol_10 > 0.05 else 1.8
            elif diff <= 0.3:
                return 0.9 if vol_10 > 0.05 else 1.9
            elif diff <= 0.4:
                return 1.0 if vol_10 > 0.05 else 2.0
            else:
                return 1.1 if vol_10 > 0.05 else 2.1

    def get_amt(self, cur_date):
        if self.moving_average is not None:
            ma = self.moving_average[:cur_date][-1]
            vol_10 = self.vol_10[:cur_date][-1]
            val = self.arr[:cur_date][-1]
            m = self._mv_multiplier(val, ma, vol_10)
            return self.amt_bench * m
        else:
            return self.amt_bench

    def run(self):
        self.res = []

        cur_date = self.start_date
        arr_idx = 0

        self.cur_mv = [0]  # 当前基金资产市值
        self.vol = [0]  # 份额
        self.tot_inv_amt = [0]  # 总投资额
        self.trd_count = [0]  # 总投资次数
        self.redeem_amt = []
        self.redeem_date = []
        self.redeem_ret = []
        self.redeem_gain = []
        avg_cost = 0  # 平均投资成本

        self.irr_helper = ([], [])
        amt_to_trade = 0

        last_trade_mv = 0

        while cur_date < self.end_date:
            while self.arr.index[arr_idx] < cur_date and arr_idx < len(self.arr.index) - 1:
                arr_idx += 1

            if self.is_to_trade(cur_date):
                amt_to_trade += self.get_amt(cur_date)
                self.irr_helper[0].append(amt_to_trade)
                self.irr_helper[1].append(cur_date)

            if self.arr.index[arr_idx] == cur_date:
                # 交易日
                p = self.arr[arr_idx]

                # 1. 处理正常的买入工作
                if amt_to_trade > 0:
                    # update data
                    self.trd_count[-1] += 1
                    _tot_cost = sum(self.tot_inv_amt) * avg_cost
                    self.tot_inv_amt[-1] += amt_to_trade
                    self.vol[-1] += amt_to_trade * (1 - self.purchase_rate) / p
                    avg_cost = (_tot_cost + amt_to_trade * p) / sum(self.tot_inv_amt)
                    self.res.append({'dt': cur_date, 'amt_to_trade': amt_to_trade, 'cur_mv': self.cur_mv[-1],
                                     'tot_inv_amt': self.tot_inv_amt[-1], 'avg_cost': avg_cost, 'p': p})
                    last_trade_price = p
                    amt_to_trade = 0
                    # print(f'[{cur_date}] (tot){_tot_cost} (avg){avg_cost} (tot_inv_amt){str(self.tot_inv_amt)}')

                # 更新 当前基金资产市值
                self.cur_mv[-1] = self.vol[-1] * p

                # 止盈逻辑
                if self.stop_profit:
                    rate = self.cur_mv[-1] / self.tot_inv_amt[-1] - 1 if self.tot_inv_amt[-1] > 0 else 0
                    if rate >= self.stop_profit:
                        next_idx = arr_idx + 1 if arr_idx + 1 < len(self.arr) else arr_idx
                        next_p = self.arr[next_idx]
                        redeem_amt = self.vol[-1] * next_p * (1 - self.redeem_rate)
                        self.redeem_amt.append(redeem_amt)
                        self.redeem_date.append(cur_date)
                        # self.redeem_ret.append(rate)
                        self.redeem_ret.append(redeem_amt / self.tot_inv_amt[-1])
                        self.redeem_gain.append(redeem_amt - self.tot_inv_amt[-1])
                        self.cur_mv.append(0)
                        self.vol.append(0)
                        self.tot_inv_amt.append(0)
                        self.trd_count.append(0)
                        self.irr_helper[0].append(-1 * redeem_amt)
                        self.irr_helper[1].append(cur_date)
                        # irr_helper[1].append(self.arr.index[next_idx])

            cur_date += datetime.timedelta(days=1)

        # stat
        self.irr_helper[0].append(-1 * (
        self.cur_mv[-1] + amt_to_trade))  # 如果终止定投前，到了月初投入一笔前，但是一直没到交易日，没有买入资产，这笔前进入了irr_helper，但是没有进入 资产里。终止定投取钱，会少最后一笔
        self.irr_helper[1].append(self.end_date)
        self.irr = xirr(self.irr_helper[0], self.irr_helper[1])
        self.tot_mv = self.cur_mv[-1] + sum(self.redeem_amt)
        self.tot_ret = (self.tot_mv) / sum(self.tot_inv_amt) - 1
        self.ret = self.convert_ret(self.tot_ret)
        self.tot_gain = self.tot_mv - sum(self.tot_inv_amt)
        # print(f'(irr){self.irr} (cur_mv){self.tot_mv} (tot_inv_amt){sum(self.tot_inv_amt)} (ret){self.ret} (cnt){sum(self.trd_count)} (tot_ret){self.tot_ret} (tot_gain){self.tot_gain}')
        '''
        if self.stop_profit:
            print('cur_mv', self.cur_mv)
            print('tot_inv_amt', self.tot_inv_amt)
            for i in range(0, len(self.redeem_amt)):
                print(f'{self.redeem_date[i]} (ret){self.redeem_ret[i]} (inv){self.tot_inv_amt[i]} (gain){self.redeem_gain[i]} (amt){self.redeem_amt[i]}')
        '''

    def get_irr(self):
        # self.res = []
        cur_date = self.start_date
        arr_idx = 0
        self.cur_mv = 0  # 当前基金资产市值
        self.vol = 0  # 份额
        self.tot_inv_amt = 0  # 总投资额
        self.irr_helper = ([], [])
        amt_to_trade = 0
        while cur_date < self.end_date:
            while self.arr.index[arr_idx] < cur_date and arr_idx < len(self.arr.index) - 1:
                arr_idx += 1

            if self.is_to_trade(cur_date):
                amt_to_trade += self.get_amt(cur_date)
                self.irr_helper[0].append(amt_to_trade)
                self.irr_helper[1].append(cur_date)

            if self.arr.index[arr_idx] == cur_date:
                # 交易日
                p = self.arr[arr_idx]

                # 1. 处理正常的买入工作
                if amt_to_trade > 0:
                    self.tot_inv_amt += amt_to_trade
                    self.vol += amt_to_trade * (1 - self.purchase_rate) / p
                    amt_to_trade = 0

                # 更新 当前基金资产市值
                self.cur_mv = self.vol * p

                # 止盈逻辑
                if self.stop_profit:
                    rate = self.cur_mv / self.tot_inv_amt - 1 if self.tot_inv_amt > 0 else 0
                    if rate >= self.stop_profit:
                        next_idx = arr_idx + 1 if arr_idx + 1 < len(self.arr) else arr_idx
                        next_p = self.arr[next_idx]
                        redeem_amt = self.vol * next_p * (1 - self.redeem_rate)
                        self.cur_mv = 0
                        self.vol = 0
                        self.tot_inv_amt = 0
                        self.irr_helper[0].append(-1 * redeem_amt)
                        self.irr_helper[1].append(cur_date)

            cur_date += datetime.timedelta(days=1)

        # stat
        self.irr_helper[0].append(-1 * (self.cur_mv + amt_to_trade))
        self.irr_helper[1].append(self.end_date)
        self.irr = xirr(self.irr_helper[0], self.irr_helper[1])
        return self.irr

    def get_aip_total_return(self):
        cash_flow = self.irr_helper[0]
        cash_inputs = [i for i in cash_flow if i > 0]
        cash_outputs = [i for i in cash_flow if i < 0]
        return (-sum(cash_outputs) / sum(cash_inputs)) - 1

    def get_fund_ret(self):
        _arr = self.arr[self.start_date:self.end_date]
        return _arr[-1] / _arr[0] - 1

    def get_fund_vol(self):
        _arr = self.arr[self.start_date:self.end_date]
        return _arr.pct_change(1).std(ddof=1)

    def get_cash_flow(self):
        dic = {'datetime': self.irr_helper[1], 'cash_flow': self.irr_helper[0]}
        return pd.DataFrame(dic)

    def plot(self):
        df_net_arr = pd.DataFrame(self.res).set_index('dt')[['p', 'avg_cost']]
        df_net_arr.plot.line(figsize=(12, 6))
        plt.legend(fontsize=15)
        plt.title('price vs avg_cost', fontsize=25)
        plt.show()


class MultipleAIPCalculator(object):

    def __init__(self, arr, start_date, end_date, stop_profit=None, strategy=None,
                 amt_bench=1000, mode='1w', mode_offset=0, purchase_rate=0.001, redeem_rate=0.001, weight_list=None, is_set_group_amt=False):
        # weight_list = 【30，40，30】权重
        self.arr = [arr.dropna()] if not isinstance(arr, list) else [i.dropna() for i in arr]
        self.weight_list = [1 for i in self.arr ] if weight_list is None else weight_list
        _sum = sum(self.weight_list)
        self.weight_list = [i/_sum for i in self.weight_list]
        self.start_date = max([start_date] + [i.index[0] for i in self.arr])
        self.end_date = end_date or datetime.datetime.now().date()
        self.amt_bench = amt_bench
        self.mode = mode
        self.mode_offset = mode_offset
        self.yearly_multiplier = self.get_yearly_multiplier(mode)
        self.purchase_rate = purchase_rate if isinstance(purchase_rate, list) else [purchase_rate]
        self.redeem_rate = redeem_rate if isinstance(redeem_rate, list) else [redeem_rate]
        self.stop_profit = stop_profit
        self.is_set_group_amt = is_set_group_amt
        # calc
        self.last_trade_day = None
        self.moving_average = None
        if strategy:
            if strategy.startswith('ma'):
                window_size = int(strategy[2:])
                self.moving_average = [arr_i.rolling(window_size).mean().fillna(0) for arr_i in self.arr]
                self.vol_10 = [arr_i.pct_change(1).rolling(10).std().fillna(0) for arr_i in self.arr]
                
    def get_yearly_multiplier(self, mode):
        m = None
        if mode == '1d':
            m = 365
        elif mode == '1w':
            m = 52
        elif mode == '2w':
            m = 26
        elif mode == '1m':
            m = 12
        else:
            raise Exception(f'unexpected trade mode: {mode}')
        return m

    # 判断是否当前要进行交易
    def is_to_trade(self, dt):
        if self.mode == '1d':
            return True
        elif self.mode == '1w':
            return dt.weekday() == self.mode_offset
        elif self.mode == '2w':
            return dt.weekday() == self.mode_offset and (self.last_trade_day is None or (dt - self.last_trade_day).days > 7)
        elif self.mode == '1m':
            return dt.day == self.mode_offset + 1
        else:
            raise Exception('unexpected trade mode: ' + self.mode)
    
    def convert_irr(self, irr):
        return pow(irr + 1, self.yearly_multiplier) - 1
        
    def convert_ret(self, ret):
        days = (self.end_date - self.start_date).days
        return pow(ret + 1, 365 / days) - 1

    def _mv_multiplier(self, val, ma, vol_10):
        if ma == 0:
            return 1
        diff = (val - ma) / ma
        if diff > 0:
            if diff <= 0.15:
                return 0.9
            elif diff <= 0.5:
                return 0.8
            elif diff <= 1:
                return 0.7
            else:
                return 0.6
        else:
            diff = abs(diff)
            if diff <= 0.05:
                return 0.6 if vol_10 > 0.05 else 1.6
            elif diff <= 0.1:
                return 0.7 if vol_10 > 0.05 else 1.7
            elif diff <= 0.2:
                return 0.8 if vol_10 > 0.05 else 1.8
            elif diff <= 0.3:
                return 0.9 if vol_10 > 0.05 else 1.9
            elif diff <= 0.4:
                return 1.0 if vol_10 > 0.05 else 2.0
            else:
                return 1.1 if vol_10 > 0.05 else 2.1

    def get_amt(self, cur_date, idx=0):
        if self.moving_average is not None:
            ma = self.moving_average[idx][:cur_date][-1]
            vol_10 = self.vol_10[idx][:cur_date][-1]
            val = self.arr[idx][:cur_date][-1]
            m = self._mv_multiplier(val, ma, vol_10)
            return self.amt_bench * m
        else:
            return self.amt_bench
    
    def get_irr(self):
        #self.res = []
        cur_date = self.start_date
        arr_idx = 0
        self.cur_mv = [[0] for i in self.arr] # 当前基金资产市值
        _tmp_amount = [0 for i in self.arr] # 当期各资产投资量

        self.vol = [0 for i in self.arr] # 份额
        self.inv_amt = [[0] for i in self.arr] # 总投资额
        self.max_cur_inv_amt = [0 for i in self.arr]  # 当期之前最大投入
        self.avg_cost = [0 for i in self.arr] # 平均花费
        self.trd_count = [0]  # 总投资次数

        self.redeem_amt = [[] for i in self.arr]  # 赎回金额
        self.redeem_inv_amt = [[] for i in self.arr]  # 赎回时投入
        self.redeem_date = [[] for i in self.arr]  # 赎回日期
        self.redeem_ret = [[] for i in self.arr]  # 赎回比率
        self.redeem_gain = [[] for i in self.arr]  # 赎回收益

        self.irr_helper = [([], []) for i in self.arr]
        self.res = [[] for i in self.arr]
        amt_to_trade = [0 for i in self.arr]
        dts = pd.DataFrame(self.arr).T.index
        dts = dts[(dts >= self.start_date) & (dts <= self.end_date)]
        while cur_date < self.end_date:
            while dts[arr_idx] < cur_date and arr_idx < len(dts) - 1:
                arr_idx += 1

            if self.is_to_trade(cur_date):
                for idx in range(len(self.arr)):
                    _tmp_amount[idx] = self.get_amt(cur_date, idx) * self.weight_list[idx]
                #if rebalance
                if self.is_set_group_amt:
                    _sum = sum(_tmp_amount)
                    _rate = self.amt_bench/_sum if _sum !=0 else 1
                    _tmp_amount = [i * _rate for i in _tmp_amount ]
                    
                for idx in range(len(self.arr)):
                    self.irr_helper[idx][0].append(_tmp_amount[idx])
                    amt_to_trade[idx] += _tmp_amount[idx]
                    self.irr_helper[idx][1].append(cur_date)
                
            if dts[arr_idx] == cur_date:
                # 交易日
                for idx, arr in enumerate(self.arr):
                    p = self.arr[idx][cur_date]
                    # 1. 处理正常的买入工作
                    if amt_to_trade[idx] > 0:
                        self.trd_count[-1] += 1
                        self.inv_amt[idx][-1] += amt_to_trade[idx]
                        # _tot_cost = sum(self.inv_amt[idx]) * self.avg_cost[idx]
                        self.vol[idx] += amt_to_trade[idx] * (1 - self.purchase_rate[idx]) / p
                        self.avg_cost[idx] = self.inv_amt[idx][-1] / self.vol[idx]
                        self.cur_mv[idx][-1] = self.vol[idx] * p
                        tot_cur_mv = sum(self.cur_mv[idx])
                        tot_cur_inv_mv = sum(self.inv_amt[idx])
                        self.res[idx].append({
                            'dt': cur_date,
                            'amt_to_trade': amt_to_trade,
                            'cur_mv': self.cur_mv[idx][-1],
                            'inv_amt': tot_cur_inv_mv,
                            'avg_cost': self.avg_cost[idx],
                            'cur_gain': tot_cur_mv - tot_cur_inv_mv,
                            'max_cur_inv_amt': max(self.max_cur_inv_amt[idx], max(self.inv_amt[idx])),
                            'p': p,
                        })
                        #print(f'dt{cur_date} p {p} amt {amt_to_trade[idx]} v {self.vol[idx]} ')

                        # 交易完成 将各资金池制空
                        amt_to_trade[idx] = 0

                    # 更新 当前基金资产市值
                    self.cur_mv[idx][-1] = self.vol[idx] * p
                    #print(f'dt {cur_date} mv {self.cur_mv[idx]} vol {self.vol[idx]} p {p} ')
                # 止盈逻辑
                if self.stop_profit:
                    tot_mv = sum([i[-1] for i in self.cur_mv])
                    tot_inv_amt = sum([i[-1] for i in self.inv_amt])
                    rate = tot_mv / tot_inv_amt - 1 if tot_inv_amt > 0 else 0
                    if rate >= self.stop_profit:
                        for idx in range(len(self.arr)):
                            next_idx = arr_idx + 1 if arr_idx + 1 < len(dts) else arr_idx
                            next_p = self.arr[idx][dts[next_idx]]
                            redeem_amt = self.vol[idx] * next_p * (1 - self.redeem_rate[idx])

                            self.redeem_amt[idx].append(redeem_amt)
                            self.redeem_inv_amt[idx].append(self.inv_amt[idx][-1])
                            self.redeem_date[idx].append(cur_date)
                            self.redeem_ret[idx].append(redeem_amt / self.inv_amt[idx][-1])
                            self.redeem_gain[idx].append(redeem_amt - self.inv_amt[idx][-1])

                            self.cur_mv[idx].append(0)
                            self.vol[idx]=0
                            self.inv_amt[idx].append(0)

                            # todo 这里处理了一个日期有两条数据的问题，可以看看是否有更好的处理方式
                            if self.irr_helper[idx][1][-1] == cur_date:
                                self.irr_helper[idx][0][-1] -= redeem_amt
                            else:
                                self.irr_helper[idx][0].append(-1 * redeem_amt)
                                self.irr_helper[idx][1].append(cur_date)

            cur_date += datetime.timedelta(days=1)
        # stat
        res = []
        for idx in range(len(self.arr)):
            self.irr_helper[idx][0].append(-1 * (self.cur_mv[idx][-1] + amt_to_trade[idx]))
            self.irr_helper[idx][1].append(self.end_date)
            i = self.irr_helper[idx]
            dic = {'datetime': i[1], idx: i[0]}
            df = pd.DataFrame(dic).set_index('datetime')
            res.append(df)
        self.cash_flow = pd.concat(res, axis=1, sort=True)
        self.cash_flow_sum = self.cash_flow.sum(axis=1)
        self.irr = xirr(self.cash_flow_sum.values, self.cash_flow_sum.index)

        # 非最后一天的负数是止盈赎回
        self.tot_mv = sum([self.cur_mv[idx][-1] + sum(self.redeem_amt[idx]) for idx in range(len(self.arr))])
        self.tot_inv_amt = sum( [sum(i) for i in self.inv_amt])
        self.tot_ret = (self.tot_mv) / self.tot_inv_amt - 1
        self.ret = self.convert_ret(self.tot_ret)
        self.tot_gain = self.tot_mv - self.tot_inv_amt
        return self.irr

    def get_funds_tot_inv_amt(self):
        return [sum(i) for i in self.inv_amt]

    def get_funds_tot_mv(self):
        return [self.cur_mv[idx][-1] + sum(self.redeem_amt[idx]) for idx in range(len(self.arr))]

    def get_funds_tot_gain(self):
        x = self.get_funds_tot_inv_amt()
        y = self.get_funds_tot_mv()
        return [y - x for x, y in zip(x, y )]

    def get_funds_tot_ret(self):
        x = self.get_funds_tot_inv_amt()
        y = self.get_funds_tot_mv()
        return [y / x - 1 for x, y in zip(x, y )]

    def get_funds_tot_irr(self):
        res = []
        for idx in range(len(self.arr)):
            res.append(xirr(
                self.irr_helper[idx][0],
                self.irr_helper[idx][1],
            ))
        return res

    def get_aip_total_return(self):
        res = []
        for i in self.irr_helper:
            cash_flow = i[0]
            cash_inputs = [i for i in cash_flow if i > 0]
            cash_outputs = [i for i in cash_flow if i < 0]
            res.append((-sum(cash_outputs) / sum(cash_inputs)) - 1)
        return res

    def get_fund_ret(self):
        res = []
        for i in self.arr:
            _arr = i[self.start_date:self.end_date]
            res.append(_arr[-1] / _arr[0] - 1)
        return res

    def get_fund_vol(self):
        res = []
        for i in self.arr:
            _arr = i[self.start_date:self.end_date]
            res.append(_arr.pct_change(1).std(ddof=1))
        return res

    def get_cash_flow(self):
        dic = {'datetime':self.irr_sum, 'cash_flow':self.irr_helper[0]}
        return pd.DataFrame(dic)

    def plot(self):
        df_net_arr = pd.DataFrame(self.res).set_index('dt')[['p', 'avg_cost']]
        df_net_arr.plot.line(figsize=(12,6))
        plt.legend(fontsize=15)
        plt.title('price vs avg_cost',fontsize=25)
        plt.show()

if __name__ == "__main__":
    pass
    '''
    from surfing.resource.data_store import *
    m = DataStore.load_dm()
    b_d = datetime.date(2015,1,5)
    e_d = datetime.date(2019,2,5)
    index_id_1 = 'hs300'
    index_id_2 = 'csi500'
    index_id_3 = 'gem'
    price1 = index_price[index_id_1]
    price2 = index_price[index_id_2]
    price3 = index_price[index_id_3]
    price = [price1, price2, price3]
    amt_bench = 1000
    mode = '1m'
    mode_offset = 0
    purchase_rate = [0,0,0]#index_fee_avg.AssetPurchaseRate[index_id]
    redeem_rate = [0,0,0]#index_fee_avg.AssetRedeemRate[index_id]
    weight_list = [30,40,30]
    self = AIPCalculator(price, b_d, e_d, None, None, amt_bench, mode, mode_offset, purchase_rate, redeem_rate, weight_list)
    self.get_irr()
    '''


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#########################################################################
# Author: jonyqin
# Created Time: Thu 11 Sep 2014 01:53:58 PM CST
# File Name: ierror.py
# Description:定义错误码含义 
#########################################################################
WXBizMsgCrypt_OK = 0
WXBizMsgCrypt_ValidateSignature_Error = -40001
WXBizMsgCrypt_ParseXml_Error = -40002
WXBizMsgCrypt_ComputeSignature_Error = -40003
WXBizMsgCrypt_IllegalAesKey = -40004
WXBizMsgCrypt_ValidateAppid_Error = -40005    # Used in wechat
WXBizMsgCrypt_ValidateCorpid_Error = -40005   # Used in wxwork
WXBizMsgCrypt_EncryptAES_Error = -40006
WXBizMsgCrypt_DecryptAES_Error = -40007
WXBizMsgCrypt_IllegalBuffer = -40008
WXBizMsgCrypt_EncodeBase64_Error = -40009
WXBizMsgCrypt_DecodeBase64_Error = -40010
WXBizMsgCrypt_GenReturnXml_Error = -40011




from typing import Dict
import requests
import json
from .config import SurfingConfigurator


class WechatBot(object):

    def __init__(self):
        self.wechat_webhook = SurfingConfigurator().get_wechat_webhook_settings('wechat_webhook')

    def send(self, markdown_content):
        print(markdown_content)
        # Send markdown content
        message = {
            'msgtype': 'markdown',
            'markdown': {
                'content': markdown_content
            }
        }
        res = requests.post(url=self.wechat_webhook, data=json.dumps(message), timeout=20)

    def send_data_update_status(self, dt, failed_tasks, updated_count):
        # Send markdown content
        markdown_content = f'{dt} 已完成数据更新，'
        if len(failed_tasks) == 0:
            markdown_content += '所有数据更新成功'
        else:
            markdown_content += '以下数据更新失败: '
            for key, task_names in failed_tasks.items():
                if len(task_names) > 0:
                    markdown_content += f'\n>{key}: <font color=\"error\">{task_names}</font>'

        markdown_content += '\n更新数据统计: '
        for key, item_count in updated_count.items():
            for item, count in item_count.items():
                markdown_content += f'\n>{key}.{item}: <font color=\"info\">{count}</font>'

        return self.send(markdown_content)

    def send_new_fund_list(self, dt, new_funds: Dict[str, str], new_delisted_funds: Dict[str, str]):
        markdown_content = f'#### {dt}'
        if new_funds is None or new_delisted_funds is None:
            markdown_content += f' **<font color=\"warning\">获取今日新基金或摘牌基金列表失败</font>**'
        else:
            if new_funds:
                markdown_content += f'\n>新基金({len(new_funds)}): <font color=\"info\">{new_funds}</font>'
            else:
                markdown_content += f'\n>无新基金'
            if new_delisted_funds:
                markdown_content += f'\n\n>摘牌基金({len(new_delisted_funds)}): <font color=\"info\">{new_delisted_funds}</font>'
            else:
                markdown_content += f'\n\n>无摘牌基金'
        self.send(markdown_content)




import datetime
import traceback
import json
import numpy as np
import pandas as pd
from typing import List, Dict, Optional, Tuple, Union

from ....constant import SectorType, IndexPriceSource, INDEX_PRICE_EXTRA_TRADE_DAYS
from ...api.raw import RawDataApi
from ...api.basic import BasicDataApi
from ...view.basic_models import *
from .basic_data_helper import BasicDataHelper


class BasicDataPart1:

    def __init__(self, data_helper: BasicDataHelper):
        self._data_helper = data_helper
        self._raw_data_api = RawDataApi()
        self._basic_data_api = BasicDataApi()
        self._index_info_df = self._basic_data_api.get_index_info()
        self._fund_info_df = self._basic_data_api.get_fund_info()

    def _status_mapping(self, x):
        if x == 'Open':
            return 0
        elif x == 'Suspended':
            return 1
        elif x == 'Limited':
            return 2
        elif x == 'Close':
            return 3
        else:
            return None

    def _find_tag(self, symbol, wind_class_II):
        if '沪深300' in symbol and wind_class_II in ['普通股票型基金', '增强指数型基金', '被动指数型基金']:
            return 'A股大盘'
        elif '中证500' in symbol and wind_class_II in ['普通股票型基金', '增强指数型基金', '被动指数型基金']:
            return 'A股中盘'
        elif '标普500' in symbol:
            return '美股大盘'
        elif '创业板' in symbol and wind_class_II in ['普通股票型基金', '增强指数型基金', '被动指数型基金']:
            return '创业板'
        elif '德国' in symbol:
            return '德国大盘'
        elif '日本' in symbol or '日经' in symbol:
            return '日本大盘'
        elif (('国债' in symbol or '利率' in symbol or '金融债' in symbol)
              and wind_class_II in ['短期纯债型基金', '中长期纯债型基金', '被动指数型债券基金']):
            return '利率债'
        elif (('信用' in symbol or '企债' in symbol or '企业债' in symbol)
              and wind_class_II in ['短期纯债型基金', '中长期纯债型基金', '被动指数型债券基金']):
            return '信用债'
        elif '黄金' in symbol:
            return '黄金'
        elif '原油' in symbol or '石油' in symbol or '油气' in symbol:
            return '原油'
        elif ('地产' in symbol or '金融' in symbol) and ('美国'not in symbol):
            return '房地产'
        else:
            return 'null'

    def fund_info(self):
        # Update manually
        # Not verified
        track_index_df = pd.read_csv('./data/fund_track_index.csv', index_col=0)
        wind_fund_info = self._raw_data_api.get_wind_fund_info()
        fund_fee = self._raw_data_api.get_fund_fee()
        wind_fund_info['order_book_id'] = [_.split('!')[0].split('.')[0] for _ in wind_fund_info['wind_id'] ]
        res = []
        for i in wind_fund_info['wind_id']:
            if not '!' in i:
                res.append(0)
            else:
                res_i = int(i.split('!')[1].split('.')[0])
                res.append(res_i)
        wind_fund_info['transition'] = res
        wind_fund_info['fund_id'] = [o + '!' + str(t) for o, t in zip(wind_fund_info['order_book_id'], wind_fund_info['transition'])]
        wind_fund_info = wind_fund_info.set_index('fund_id')
        fund_fee = fund_fee.drop(['id','desc_name'], axis = 1).set_index('fund_id')
        wind_fund_info = wind_fund_info.join(fund_fee)
        wind_fund_info['fund_id'] = wind_fund_info.index
        wind_fund_info['asset_type'] = [self._find_tag(symbol, wind_class_II) for symbol, wind_class_II in zip(wind_fund_info['desc_name'],wind_fund_info['wind_class_2'])]
        wind_fund_info = wind_fund_info[[i.split('.')[1] == 'OF' for i in wind_fund_info['wind_id']]]
        wind_fund_info = wind_fund_info.set_index('wind_id')
        dic = {k:v for k,v in zip(self._index_info_df['desc_name'], self._index_info_df['index_id'])}
        wind_fund_info = wind_fund_info.join(track_index_df[['track_index']])
        wind_fund_info['wind_id'] = wind_fund_info.index
        wind_fund_info['track_index'] = wind_fund_info['track_index'].map(lambda x: dic.get(x,'none'))
        self._data_helper._upload_basic(wind_fund_info, FundInfo.__table__.name)

    def index_info(self):
        # Update manually
        # Not verified
        df = pd.read_csv('./data/index_info.csv')
        self._data_helper._upload_basic(df, IndexInfo.__table__.name)

    def index_info_from_raw(self, index_list: Union[Tuple[str], List[str]], index_ids: List[str]):
        assert len(index_list) == len(index_ids), 'index list length should match the ids'

        df = self._raw_data_api.get_em_index_info(index_list).drop(columns=['name', 'publish_date', '_update_time'])
        assert df.shape[0] == len(index_ids), 'index info length should match the ids'
        df = df.rename(columns={'short_name': 'desc_name'})
        # 先全体置为false，后续可以手工修改
        df['is_select'] = False
        df['tag_method'] = 'PE百分位'
        df['price_source'] = df['price_source'].transform(lambda x: x.name)
        # index_id是人工起的，所以只能作为参数传进来
        df['index_id'] = index_ids
        self._data_helper._upload_basic(df, IndexInfo.__table__.name)

    @staticmethod
    def _lambda_func_index_component_2(em_id_to_index_id: Dict[str, str], x: str):
        info = json.loads(x)
        r = []
        for one in info:
            em_id = one[0]
            if em_id in ['000915.SH', '000916.SH']:
                em_id = em_id.split('.')[0] + '.CSI'
            r.append((em_id_to_index_id[em_id], one[1]))
        return json.dumps(r)

    @staticmethod
    def _lambda_func_index_component_3(wind_id_to_id: Dict[str, str], x: str):
        if x is None:
            return x
        r = []
        info = json.loads(x)
        for one in info:
            try:
                r.append(wind_id_to_id[one + '.OF'])
            except KeyError:
                r.append(one + '!0')
        return json.dumps(r)

    def index_component(self):
        index_info_df = self._index_info_df[self._index_info_df.em_id.notna()]
        index_info_df = index_info_df.set_index('em_id')
        em_id_to_index_id: Dict[str, str] = index_info_df['index_id'].to_dict()

        df = self._raw_data_api.get_cs_index_component().drop(columns=['_update_time'])
        df['index_id'] = df['index_id'].map(em_id_to_index_id)
        df['sector'] = df['sector'].transform(lambda x: BasicDataPart1._lambda_func_index_component_2(em_id_to_index_id, x))
        df['id_cat'] = df['id_cat'].transform(lambda x: x.name)
        fund_info_df = self._fund_info_df.set_index('wind_id')
        wind_id_to_id: Dict[str, str] = fund_info_df['fund_id'].to_dict()
        df['related_funds'] = df['related_funds'].transform(lambda x: BasicDataPart1._lambda_func_index_component_3(wind_id_to_id, x))
        self._data_helper._upload_basic(df, IndexComponent.__table__.name)

    def fund_nav_from_rq(self, start_date, end_date):
        df = self._raw_data_api.get_rq_fund_nav(start_date, end_date)
        df['fund_id'] = df.apply(
            lambda x: self._data_helper._get_fund_id_from_order_book_id(x['order_book_id'], x['datetime']), axis=1)
        df['subscribe_status'] = df['subscribe_status'].map(self._status_mapping)
        df['redeem_status'] = df['redeem_status'].map(self._status_mapping)
        df = df[df['fund_id'].notna()]
        self._data_helper._upload_basic(df, FundNav.__table__.name)

    def fund_nav(self, start_date, end_date):
        try:
            start_date = datetime.datetime.strptime(start_date, '%Y%m%d') - datetime.timedelta(days=5)
            del_date = start_date + datetime.timedelta(days=1)
            start_date = start_date.strftime('%Y%m%d')
            df = self._raw_data_api.get_em_fund_nav(start_date, end_date)
            df['fund_id'] = df.apply(
                lambda x: self._data_helper._get_fund_id_from_order_book_id(x['CODES'].split('.')[0], x['DATES']), axis=1)
            df = df.drop(['CODES'], axis=1)
            df = df[df['fund_id'].notna()]
            df = df.rename(columns={
                    'DATES': 'datetime',
                    'ORIGINALUNIT': 'unit_net_value',
                    'ORIGINALNAVACCUM': 'acc_net_value',
                    'ADJUSTEDNAV': 'adjusted_net_value',
                    'UNITYIELD10K': 'daily_profit',
                    'YIELDOF7DAYS': 'weekly_yield'
                })
            df['weekly_yield'] = df['weekly_yield'].map(lambda x: x / 100.0)

            anv_df = df.pivot_table(index='datetime', columns='fund_id', values='adjusted_net_value').fillna(method='ffill')
            df = df.set_index(['datetime', 'fund_id'])
            df['change_rate'] = (anv_df / anv_df.shift(1) - 1).stack()

            end_date = datetime.datetime.strptime(end_date, '%Y%m%d').date()
            BasicDataApi().delete_fund_nav(del_date, end_date)
            df = df[df.index.get_level_values(level='datetime').date >= del_date.date()]
            self._data_helper._upload_basic(df.reset_index(), FundNav.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def fund_benchmark(self):
        df = pd.read_csv('./data/benchmark_res2.csv')
        df = df[['fund_id', 'fund_em_id', 'benchmark', 'index_text', 'assets', 'industry']].rename(columns={'fund_em_id': 'em_id'})
        self._data_helper._upload_basic(df, FundBenchmark.__table__.name)

    def index_price(self, start_date, end_date):
        try:
            # 多取N个交易日的数据
            start_date = datetime.datetime.strptime(start_date, '%Y%m%d') - datetime.timedelta(days=90)
            start_date = start_date.strftime('%Y%m%d')
            trading_day_df = BasicDataApi().get_trading_day_list(start_date, end_date)
            assert trading_day_df.datetime.size >= INDEX_PRICE_EXTRA_TRADE_DAYS, f'should at least have {INDEX_PRICE_EXTRA_TRADE_DAYS} trading days'
            # del_date在start_date后边一天，确保可以正确计算ret等daily difference
            start_date = trading_day_df.datetime.array[-INDEX_PRICE_EXTRA_TRADE_DAYS]
            del_date = trading_day_df.datetime.array[-(INDEX_PRICE_EXTRA_TRADE_DAYS-1)]
            print(f'[index_price] (start_date){start_date} (del_date){del_date} (end_date){end_date}')
            # 准备多个数据源
            cm_index_price = self._raw_data_api.get_raw_cm_index_price_df(start_date, end_date).drop(columns=['_update_time']).set_index('datetime').fillna(method='ffill')
            cxindex_index_price = self._raw_data_api.get_cxindex_index_price_df(start_date, end_date).drop(columns=['_update_time'])
            yahoo_index_price = self._raw_data_api.get_yahoo_index_price_df(start_date, end_date).drop(columns=['_update_time'])
            index_info_df = self._index_info_df[self._index_info_df.price_source == IndexPriceSource.default]
            em_id_list: Dict[str, str] = index_info_df[['em_id', 'index_id']].dropna().set_index('em_id').to_dict()['index_id']
            em_index_price = self._raw_data_api.get_em_index_price(start_date, end_date, em_id_list.keys()).drop(columns=['_update_time'])
            df_list = []
            # 处理Choice数据
            index_id_list = []
            for em_id, index_id in em_id_list.items():
                df = em_index_price[em_index_price['em_id'] == em_id]
                df = df.drop(columns=['em_id'])
                df['index_id'] = index_id
                df['ret'] = df.close / df.close.shift(1) - 1
                df = df[df.datetime >= pd.to_datetime(del_date).date()]
                df_list.append(df)
                if not df.empty:
                    index_id_list.append(index_id)

            end_date_dt = pd.to_datetime(end_date).date()
            index_list = ['sp500', 'dax30', 'n225']
            cm_index_list = ['sp500rmb', 'dax30rmb', 'n225rmb']
            for i, c in zip(index_list, cm_index_list):
                df = yahoo_index_price[yahoo_index_price['index_id'] == i]
                df = df.set_index('datetime').reindex(cm_index_price.index).fillna(method='ffill')
                # 存进去人民币计价的数据
                df['close'] *= cm_index_price['usd_central_parity_rate']
                df = df[['close']]
                df['index_id'] = c
                df['ret'] = df.close / df.close.shift(1) - 1
                df = df.loc[[end_date_dt], :]
                df_list.append(df.reset_index())

            df_list.append(cxindex_index_price[cxindex_index_price.datetime == end_date_dt])
            # 存进去美元计价的数据
            df_list.append(yahoo_index_price[yahoo_index_price.datetime == end_date_dt])

            df = pd.concat(df_list, ignore_index=True)
            # TODO: 改成原子提交下边两个操作
            BasicDataApi().delete_index_price(index_id_list, del_date, end_date)
            self._data_helper._upload_basic(df.replace([np.Inf, -np.Inf], np.nan), IndexPrice.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def fund_rating_latest(self):
        try:
            fund_rating_df = self._raw_data_api.get_fund_rating()
            fund_info = self._fund_info_df[['fund_id', 'order_book_id', 'start_date', 'end_date']]
            score_name = ['zs', 'sh3', 'sh5', 'jajx']
            score_dic = {}
            for s in score_name:
                score_df = fund_rating_df[['order_book_id', 'datetime', s]]
                datelist = sorted(list(set(score_df.dropna().datetime.tolist())))
                score_dic[s] = datelist[-1]
            res = []
            for s, d in score_dic.items():
                try:
                    df = fund_rating_df[['order_book_id', 'datetime', s]]
                    df = df[df['datetime'] == d]
                    con1 = fund_info['start_date'] <= d
                    con2 = fund_info['end_date'] >= d
                    fund_info_i = fund_info[con1 & con2]
                    dic = {row['order_book_id']: row['fund_id']
                        for index, row in fund_info_i.iterrows()}
                    df['fund_id'] = df['order_book_id'].map(lambda x: dic[x])
                    df = df[['fund_id', s]].copy().set_index('fund_id')
                    res.append(df)
                except Exception as e:
                    print(e)
            df = pd.concat(res, axis=1, sort=False)
            df['fund_id'] = df.index
            df['update_time'] = datetime.date.today()
            self._data_helper._upload_basic(df, FundRatingLatest.__table__.name, to_truncate=True)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def rq_stock_price(self, start_date, end_date):
        try:
            stock_price_df = self._raw_data_api.get_rq_stock_price(start_date, end_date)
            stock_post_price_df = self._raw_data_api.get_rq_stock_post_price(start_date, end_date)

            stock_post_price_df['adj_close'] = stock_post_price_df['close']
            stock_post_price_df = stock_post_price_df.filter(items=['adj_close', 'datetime', 'order_book_id'],
                axis='columns')

            stock_price_merge_df = stock_price_df.merge(stock_post_price_df, how='left', on=['datetime', 'order_book_id'])
            stock_price_merge_df['post_adj_factor'] = stock_price_merge_df['adj_close'] / stock_price_merge_df['close']
            stock_price_merge_df = stock_price_merge_df.rename(columns={'order_book_id': 'stock_id'})
            self._data_helper._upload_basic(stock_price_merge_df, StockPrice.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def rq_fund_ret(self, start_date, end_date):
        try:
            column_rename_dict = {
                'sharp_ratio': 'sharpe_ratio',
                'last_week_return': 'w1_ret',
                'last_month_return': 'm1_ret',
                'last_three_month_return': 'm3_ret',
                'last_six_month_return': 'm6_ret',
                'last_twelve_month_return': 'y1_ret',
                'max_drop_down': 'mdd',
                'annualized_returns': 'annual_ret',
                'average_size': 'avg_size',
                'information_ratio': 'info_ratio',
                'to_date_return':'to_date_ret'
            }
            column_drop_list = ['order_book_id', 'year_to_date_return', 'annualized_risk']

            df = self._raw_data_api.get_rq_fund_indicator(start_date, end_date)
            df['fund_id'] = df.apply(
                lambda x: self._data_helper._get_fund_id_from_order_book_id(x['order_book_id'], x['datetime']), axis=1)
            df = df.rename(columns=column_rename_dict).drop(columns=column_drop_list)
            df = df[df['fund_id'].notnull()]

            self._data_helper._upload_basic(df, FundRet.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def stock_info(self):
        try:
            df = self._raw_data_api.get_stock_info()
            self._data_helper._upload_basic(df, StockInfo.__table__.name, to_truncate=True)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    @staticmethod
    def _lambda_func_index_component_1(em_id_map: Dict[str, str], x: str):
        try:
            return em_id_map[x]
        except KeyError:
            return x

    def sector_info(self):
        try:
            df = pd.read_csv('./data/sector_info.csv', index_col=0)
            df['sector_type'] = df['sector_type'].map({'industry': SectorType.industry.name, 'topic': SectorType.topic.name})
            self._data_helper._upload_basic(df, SectorInfo.__table__.name, to_truncate=True)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def sector_funds(self):
        try:
            index_info_df = self._index_info_df[self._index_info_df.em_id.notna()]
            index_info_df = index_info_df.set_index('em_id')
            em_id_list: List[str] = index_info_df.index.to_list()
            em_id_to_index_id: Dict[str, str] = index_info_df['index_id'].to_dict()

            df = pd.read_csv('./data/sector_funds.csv', index_col=0)
            df['csi_index_id'] = df['csi_index_id'].transform(lambda x: BasicDataPart1._lambda_func_index_component_1({em_id.split('.')[0]: em_id for em_id in em_id_list if em_id is not None}, x))
            df['index_id'] = df['csi_index_id'].map(em_id_to_index_id)
            self._data_helper._upload_basic(df, SectorFunds.__table__.name, to_truncate=True)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def fund_size(self, start_date, end_date):
        try:
            # Since fund_size in basic db is latest snapshot, we only use end_date as param
            df = self._raw_data_api.get_em_fund_scale(end_date=end_date)
            if df is None:
                print('failed to get em fund scale for fund size')
                return False
            # 过滤掉非.OF的基金
            df = df[df.CODES.str.contains('.OF')]
            df = df.pivot_table(index='DATES', columns='CODES', values='FUNDSCALE').fillna(method='ffill')
            df = df.iloc[[-1], :].apply(lambda x: pd.Series({'fund_id': self._data_helper._get_fund_id_from_order_book_id(x.name.split('.')[0], x.index.array[0]), 'latest_size': float(x.array[0])})).T
            df = df.reset_index(drop=True)
            df = df[df.notnull().all(axis=1)]
            self._data_helper._upload_basic(df, FundSize.__table__.name, to_truncate=True)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def fund_status_latest(self, end_date: str):
        try:
            df: Optional[pd.DataFrame] = self._raw_data_api.get_em_fund_status(end_date=end_date)
            if df is None:
                print('failed to get em fund scale for fund status latest')
                return False
            # 过滤掉非.OF的基金
            df = df[df.CODES.str.contains('.OF')]
            df = df.pivot(index='DATES', columns='CODES', values=['PURCHSTATUS', 'REDEMSTATUS', 'LPLIMIT', 'OTCLPLIMIT', 'TRADESTATUS']).fillna(method='ffill')
            df = df.iloc[[-1], :].stack().reset_index()
            df['fund_id'] = df[['DATES', 'CODES']].apply(lambda x: self._data_helper._get_fund_id_from_order_book_id(x.CODES.split('.')[0], x.DATES), axis=1)
            df = df[df.fund_id.notna()]
            df = df.drop(columns=['DATES', 'CODES']).rename(columns=lambda x: FundStatusLatest.__getattribute__(FundStatusLatest, x).name)
            self._data_helper._upload_basic(df, FundStatusLatest.__table__.name, to_truncate=True)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def fund_rate(self, end_date):
        def _count_star_history(x):
            color_change = {
                '红':'蓝',
                '蓝':'白',
                '白':'红',    
            }
            if x is None:
                return None
            if '星' in x:
                return str(x.count('★') + x.count('☆') + 1) + color_change[x[-2]] 
            else:
                return str(x.count('★') + x.count('☆')) 

        def _count_star_recent(x):
            if x is None:
                return None
            if '星' in x:
                return str(x.count('★') + x.count('☆') + 1) + x[-2]
            else:
                return str(x.count('★') + x.count('☆')) 
        try:
            df = self._raw_data_api.get_em_fund_rate(end_date)
            df['fund_id'] = df.apply( lambda x : self._data_helper._get_fund_id_from_order_book_id(x.CODES.split('.')[0], x.DATES), axis=1)
            no_star_columns = ['CODES','DATES','_update_time','fund_id']
            for col in df:
                if col in no_star_columns:
                    continue 
                df[col] = df[col].map(lambda x : _count_star_recent(x))  
            df = df.drop(['CODES','_update_time'], axis=1)
            name_dic = {
                'DATES':'datetime',
                'MKTCOMPRE3YRAT':'mkt_3y_comp',
                'MORNSTAR3YRAT':'mornstar_3y',
                'MORNSTAR5YRAT':'mornstar_5y',
                'MERCHANTS3YRAT':'merchant_3y',
                'TXCOMRATRAT':'tx_comp',
                'SHSTOCKSTAR3YCOMRAT':'sh_3y_comp',
                'SHSTOCKSTAR3YSTOCKRAT':'sh_3y_stock',
                'SHSTOCKSTAR3YTIMERAT':'sh_3y_timeret',
                'SHSTOCKSTAR3YSHARPERAT':'sh_3y_sharpe',
                'SHSTOCKSTAR5YCOMRATRAT':'sh_5y_comp',
                'SHSTOCKSTAR5YSTOCKRAT':'sh_5y_stock',
                'SHSTOCKSTAR5YTIMERAT':'sh_5y_timeret',
                'SHSTOCKSTAR5YSHARPRAT':'sh_5y_sharpe',
                'JAJXCOMRAT':'jian_comp',
                'JAJXEARNINGPOWERRAT':'jian_earn',
                'JAJXACHIEVESTABILITYRAT':'jian_stable',
                'JAJXANTIRISKRAT':'jian_risk',
                'JAJXSTOCKSELECTIONRAT':'jian_stock',
                'JAJXTIMESELECTIONRAT':'jian_timeret',
                'JAJXBENCHMARKTRACKINGRAT':'jian_track',
                'JAJXEXCESSEARNINGSRAT':'jian_alpha',
                'JAJXTOTALFEERAT':'jian_fee',
                'fund_id':'fund_id',
            }
            df = df.rename(columns = name_dic)
            df = df.dropna(subset=['fund_id','datetime'], axis=0)
            self._data_helper._upload_basic(df, FundRate.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False
            
    def trading_day_list(self, start_date, end_date, is_automatic_update=True):
        try:
            # 如果是每日自动更新，我们需要去存T+1日的数据
            if is_automatic_update:
                start_date = end_date
                end_date = ''
            df = self._raw_data_api.get_em_tradedates(start_date, end_date)
            df = df.rename(columns={'TRADEDATES': 'datetime'})
            if is_automatic_update:
                df = df.iloc[[1], :]
            self._data_helper._upload_basic(df, TradingDayList.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def fund_size_hold_rate_history(self):
        start_date = '20100101'
        end_date = '20200612'
        raw = self._raw_data_api
        basic = BasicDataApi()
        fund_holding_rate = raw.get_em_fund_holding_rate(start_date=start_date)
        fund_scale = raw.get_em_fund_scale(start_date=start_date, end_date=end_date)
        fund_holding_rate['fund_id'] = fund_holding_rate.apply(
            lambda x: self._data_helper._get_fund_id_from_order_book_id(x['CODES'].split('.')[0], x['DATES']), axis=1)
        fund_scale['fund_id'] = fund_scale.apply(
            lambda x: self._data_helper._get_fund_id_from_order_book_id(x['CODES'].split('.')[0], x['DATES']), axis=1)
        fund_holding_rate = fund_holding_rate.drop(['CODES', '_update_time'], axis=1).rename(columns={'DATES':'datetime','HOLDPERSONALHOLDINGPCT':'personal_holds','HOLDINSTIHOLDINGPCT':'institution_holds'})
        fund_scale = fund_scale.drop(['CODES','_update_time'], axis=1).rename(columns={'DATES':'datetime','FUNDSCALE':'size'})
        fund_scale = fund_scale.dropna().drop_duplicates(subset=['size','fund_id'],keep='first')
        fund_holding_rate = fund_holding_rate.dropna()
        fund_list = list(set(fund_scale.fund_id))
        res = []
        for fund_id in fund_list:
            fund_hold_i = fund_holding_rate[fund_holding_rate.fund_id == fund_id].drop('fund_id',axis=1).set_index('datetime')
            fund_scale_i = fund_scale[fund_scale.fund_id == fund_id].drop('fund_id',axis=1).set_index('datetime')
            _df = fund_scale_i.join(fund_hold_i).fillna(method='ffill').reset_index()
            _df['fund_id'] = fund_id
            res.append(_df)
        df = pd.concat(res)
        df = df.drop_duplicates(subset=['datetime','fund_id'],keep='first')
        df = df.fillna(0)
        con_list = (df.personal_holds + df.institution_holds) == 100
        df = df[con_list]
        self._data_helper._upload_basic(df, Fund_size_and_hold_rate.__table__.name)

    def fund_hold_asset(self):
        #history, update only on report date
        em_hold_asset = self._raw_data_api.get_em_fund_hold_asset()
        em_hold_asset['CODES'] = em_hold_asset.apply(lambda x : 
                            self._data_helper._get_fund_id_from_order_book_id(x.CODES.split('.')[0], x.DATES), axis=1)
        em_hold_asset = em_hold_asset.dropna(subset=['CODES']).drop('_update_time', axis=1)
        name_dict = {
            'CODES':'fund_id',
            'DATES':'datetime',
            'PRTSTOCKTONAV':'stock_nav_ratio',
            'PRTBONDTONAV':'bond_nav_ratio',
            'PRTCASHTONAV':'cash_nav_ratio',
            'PRTOTHERTONAV':'other_nav_ratio',
        }
        em_hold_asset = em_hold_asset.rename(columns=name_dict)
        self._data_helper._upload_basic(em_hold_asset, 'fund_hold_asset')
        
    def fund_hold_industry(self):
        #history, update only on report date
        em_hold_industry = self._raw_data_api.get_em_fund_hold_industry()
        em_hold_industry['CODES'] = em_hold_industry.apply(lambda x : 
                            self._data_helper._get_fund_id_from_order_book_id(x.CODES.split('.')[0], x.DATES), axis=1)
        em_hold_industry = em_hold_industry.dropna(subset=['CODES']).drop('_update_time', axis=1)
        name_dict = {
            'CODES':'fund_id',
            'DATES':'datetime',
        }
        em_hold_industry = em_hold_industry.rename(columns=name_dict)
        self._data_helper._upload_basic(em_hold_industry, 'fund_hold_industry')

    def fund_hold_stock(self):
        #history, update only on report date
        em_hold_stock = self._raw_data_api.get_em_fund_hold_stock()
        em_hold_stock['CODES'] = em_hold_stock.apply(lambda x : 
                            self._data_helper._get_fund_id_from_order_book_id(x.CODES.split('.')[0], x.DATES), axis=1)
        em_hold_stock = em_hold_stock.dropna(subset=['CODES']).drop('_update_time', axis=1)
        name_dict = {
            'CODES':'fund_id',
            'DATES':'datetime',
        }
        em_hold_stock = em_hold_stock.rename(columns=name_dict)
        self._data_helper._upload_basic(em_hold_stock, 'fund_hold_stock')

    def fund_hold_bond(self):
        #history, update only on report date
        em_hold_bond = self._raw_data_api.get_em_fund_hold_bond()
        em_hold_bond['CODES'] = em_hold_bond.apply(lambda x : 
                            self._data_helper._get_fund_id_from_order_book_id(x.CODES.split('.')[0], x.DATES), axis=1)
        em_hold_bond = em_hold_bond.dropna(subset=['CODES']).drop('_update_time', axis=1)
        name_dict = {
            'CODES':'fund_id',
            'DATES':'datetime',
        }
        em_hold_bond = em_hold_bond.rename(columns=name_dict)
        self._data_helper._upload_basic(em_hold_bond, 'fund_hold_bond')

    def process_all(self, start_date, end_date):
        failed_tasks = []
        # Depend on RQ
        # if not self.rq_fund_ret(start_date, end_date):
        #     failed_tasks.append('rq_fund_ret')

        # # Depend on RQ
        # if not self.stock_info():
        #     failed_tasks.append('stock_info')

        # Depend on RQ
        # if not self.rq_stock_price(start_date, end_date):
        #     failed_tasks.append('rq_stock_price')

        if not self.index_price(start_date, end_date):
            failed_tasks.append('index_price')

        # TODO: Depend on RQ, change to EM
        if not self.fund_rating_latest():
            failed_tasks.append('fund_rating_latest')

        if not self.fund_size(start_date, end_date):
            failed_tasks.append('fund_size')

        if not self.fund_nav(start_date, end_date):
            failed_tasks.append('fund_nav')

        if not self.trading_day_list(start_date, end_date):
            failed_tasks.append('trading_day_list')

        if not self.fund_rate(end_date):
            failed_tasks.append('fund_rate')

        if not self.fund_status_latest(end_date):
            failed_tasks.append('fund_status_latest')

        return failed_tasks

if __name__ == '__main__':
    bdp = BasicDataPart1(BasicDataHelper())
    # start_date = '20200428'
    # end_date = '20200428'
    # bdp.fund_nav(start_date, end_date)
    # bdp.index_price(start_date, end_date)
    # bdp.rq_stock_price(start_date, end_date)
    # bdp.rq_fund_ret(start_date, end_date)
    # bdp.fund_rating_latest()
    # bdp.fund_size('20200513', '20200513')
    # bdp.trading_day_list('20200124', '20200301', False)
    # bdp.fund_benchmark()
    # bdp.sector_info()
    # bdp.sector_funds()
    # bdp.index_component()




from typing import Dict, Union, Optional
import datetime
import copy
import json
import numpy as np
import pandas as pd
import os
import time

from .asset_helper import SAAHelper, TAAHelper, FAHelper
from .trader import AssetTrader, FundTrader
from .report import ReportHelper
from ...data.manager.manager_fund import FundDataManager
from ...data.manager.score import ScoreFunc, FundScoreManager
from ...data.struct import AssetWeight, AssetPrice, AssetPosition, AssetValue
from ...data.struct import FundPosition, TAAParam, AssetTradeParam, FundTradeParam, FAParam
from ...util.calculator import Calculator

class FundEngine:

    DEFAULT_CASH_VALUE = 1e8

    def __init__(self, data_manager: FundDataManager, trader, taa_params:TAAParam=None, fa_params:FAParam=None, taa_param_details:dict=None, is_old_version=False, white_list:set=None, black_list:set=None, fund_score_funcs: Optional[Dict[str, Union[ScoreFunc, str]]] = None):
        self._dm = data_manager
        self._saa_helper = SAAHelper()
        self._taa_helper = TAAHelper(taa_params=taa_params, taa_param_details=taa_param_details) if taa_params or taa_param_details else None 
        self._fa_helper = FAHelper(fa_params=fa_params) if fa_params else None
        self._report_helper = ReportHelper()
        self._trader = trader
        self._is_old_version = is_old_version
        self._fund_score_funcs: Union[Dict[str, ScoreFunc], str] = fund_score_funcs
        if self._fund_score_funcs is not None:
            for index_id, func in self._fund_score_funcs.items():
                if isinstance(func, str):
                    print(f'check score func "{func}" for {index_id}')
                    FundScoreManager.test_func(func)
        self._pending_trades = []
        # 黑白名单和全部基金列表处理 得到不允许交易基金列表， 对其中基金评分惩罚
        self.white_list = white_list
        self.black_list = black_list
        self.whole_list = set(self._dm.dts.fund_list)
        disproved_set_1 = self.whole_list - self.white_list if self.white_list else set([]) 
        disproved_set_2 = self.black_list if self.black_list else set([])
        self.disproved_set = disproved_set_1.union(disproved_set_2)
        assert not self.is_fund_backtest or self._fa_helper, 'if this is a fund backtest, fa_helper is necessary'
        if self.is_fund_backtest:
            self._trader.set_helper(self._fa_helper)

    def init(self):
        if not self._dm.inited:
            self._dm.init()
        self._saa_helper.init()
        if self._taa_helper:
            self._taa_helper.init()
        self._report_helper.init()
        self._trader.init()
        # preps
        self._prep_fund_purchase_fees = None
        self._prep_fund_redeem_fees = None
        self._prep_asset_price = None
        self._prep_fund_nav = None
        self._prep_fund_unit_nav = None
        self._prep_fund_score = None
        self._prep_fund_score_raw = None
        self._prep_target_asset_allocation = None
        self._prep_target_fund_allocation = None

    def setup(self, saa: AssetWeight):
        # setup helpers
        self._saa_helper.setup(saa)
        if self._taa_helper:
            self._taa_helper.setup(saa)
        self._report_helper.setup(saa)
        self._trader.setup(saa)
        if self.is_fund_backtest:
            self._fa_helper.setup(saa)

    @property
    def is_fund_backtest(self):
        return isinstance(self._trader, FundTrader)

    def prep_data(self, dt):
        # basic
        if not self._prep_fund_purchase_fees:
            self._prep_fund_purchase_fees = self._dm.get_fund_purchase_fees()
        if not self._prep_fund_redeem_fees:
            self._prep_fund_redeem_fees = self._dm.get_fund_redeem_fees()
        # DATA
        self._prep_asset_price = self._dm.get_index_price_data(dt)
        self._prep_fund_nav = self._dm.get_fund_nav(dt) if self.is_fund_backtest else {}
        self._prep_fund_unit_nav = self._dm.get_fund_unit_nav(dt) if self.is_fund_backtest else {}
        self._prep_fund_score, self._prep_fund_score_raw = self._dm.get_fund_scores(dt=dt, fund_score_funcs=self._fund_score_funcs) if self.is_fund_backtest else ({}, {})
        self._prep_fund_score, self._prep_fund_score_raw = self._dm.white_and_black_list_filter(self._prep_fund_score, self._prep_fund_score_raw, self.disproved_set)

        # ALLOCATION
        self._prep_target_asset_allocation = self.calc_asset_allocation(dt)
        self._prep_target_fund_allocation = self.calc_fund_allocation(dt) if self.is_fund_backtest else None

    def calc_trade(self, dt, cur_asset_position: AssetPosition, cur_fund_position: FundPosition=None):
        if not self.is_fund_backtest:
            v_asset_position, asset_trade_list = self._trader.calc_asset_trade(dt, cur_asset_position, self._prep_asset_price, self._prep_target_asset_allocation)
            return asset_trade_list
        else:
            if self._is_old_version:
                v_asset_position, asset_trade_list = self._trader.calc_asset_trade(dt, cur_asset_position, self._prep_asset_price, self._prep_target_asset_allocation)
                expired_con = self._trader.has_expired_fund(cur_fund_position, self._prep_fund_score )
                if asset_trade_list or expired_con:
                    self._report_helper.update_rebalance_detail(dt, expired_con)
                    v_fund_position, fund_trade_list = self._trader.calc_fund_trade(
                        dt, self._prep_target_fund_allocation, cur_fund_position, self._prep_fund_nav, 
                        self._prep_fund_purchase_fees, self._prep_fund_redeem_fees
                    )
                    return fund_trade_list
                else:
                    return None
            assert cur_fund_position, 'cur_fund_position should not be None in fund backtest run'
            assert self._prep_target_fund_allocation, 'target fund allocation should be prepared already'
            v_fund_position, fund_trade_list, trigger_reason = self._trader.calc_trade(
                dt, self._prep_target_fund_allocation, self._prep_target_asset_allocation,
                cur_fund_position, cur_asset_position, 
                self._prep_fund_nav, self._prep_fund_score, self._prep_asset_price
            )
            self._report_helper.update_rebalance_detail(dt, trigger_reason)
            return fund_trade_list

    def finalize_trade(self, dt, trades: list, cur_asset_position: AssetPosition, cur_fund_position: FundPosition=None, disproved_set: set={}):
        if self.is_fund_backtest:
            assert cur_fund_position, 'cur_fund_position should not be None in fund backtest run'
            self._pending_trades, traded_list = self._trader.finalize_trade(dt, trades, self._prep_asset_price, cur_asset_position, cur_fund_position, self._prep_fund_nav, self._prep_fund_unit_nav,
                self._prep_fund_purchase_fees, self._prep_fund_redeem_fees, disproved_set)
        else:
            self._pending_trades, traded_list = self._trader.finalize_trade(dt, trades, self._prep_asset_price, cur_asset_position)
        return traded_list 

    def calc_asset_allocation(self, dt):
        cur_asset_price = self._dm.get_index_price_data(dt)
        cur_saa = self._saa_helper.on_price(dt, cur_asset_price)
        if self._taa_helper:
            asset_pct = self._dm.get_index_pcts_df(dt)
            cur_taa = self._taa_helper.on_price(dt, cur_asset_price, cur_saa, asset_pct, {}, self._prep_fund_score)
        else:
            cur_taa = cur_saa
        return cur_taa

    def calc_fund_allocation(self, dt):
        return self._fa_helper.on_price(dt, self._prep_target_asset_allocation, self._prep_fund_score)

    def update_reporter(self, dt, trade_list, cur_asset_position: AssetPosition, cur_fund_position: FundPosition=None):
        self._report_helper.update(dt, cur_asset_position, self._prep_asset_price, self._pending_trades, cur_fund_position if self.is_fund_backtest else None, self._prep_fund_nav, trade_list, self._prep_fund_score, self._prep_fund_score_raw, self._prep_target_asset_allocation)

class FundBacktestEngine(FundEngine):

    def run(self, saa: AssetWeight, start_date: datetime.date=None, end_date: datetime.date=None, cash: float=None, print_time=False):
        cash = cash or self.DEFAULT_CASH_VALUE
        # init position
        self.cur_asset_position = AssetPosition(cash=cash)
        self.cur_fund_position = FundPosition(cash=cash) if self.is_fund_backtest else None

        self._pending_trades = []

        # init days
        start_date = start_date or self._dm.start_date
        end_date = end_date or self._dm.end_date
        start_date = start_date if isinstance(start_date, datetime.date) else datetime.datetime.strptime(start_date,'%Y%m%d').date()
        end_date = end_date if isinstance(end_date, datetime.date) else datetime.datetime.strptime(end_date,'%Y%m%d').date()

        # setup helpers
        self.setup(saa)

        # loop trading days
        _dts = self._dm.get_trading_days()
        dts = _dts[(_dts.datetime >= start_date) & (_dts.datetime <= end_date)].datetime # df with datetime.date
        for t in dts:
            self._run_on(t, print_time=print_time)
        # init report data
        self._report_helper.plot_init(self._dm, self._taa_helper)
        #print(self._report_helper.get_fund_stat())

    def _run_on(self, dt, print_time=False):
        _tm_start = time.time()
        # prep data
        self.prep_data(dt)
        _tm_prep_data = time.time()
        # finalize trade
        traded_list = self.finalize_trade(dt, self._pending_trades, self.cur_asset_position, self.cur_fund_position, self.disproved_set)
        _tm_finalize_trade = time.time()
        # calc trade
        trade_list = self.calc_trade(dt, self.cur_asset_position, self.cur_fund_position)
        _tm_calc_trade = time.time()
        # book trade
        self.book_trades(trade_list)
        # update
        self.update_reporter(dt, traded_list, self.cur_asset_position, self.cur_fund_position)        
        _tm_finish = time.time()

        if print_time:
            print(f'{dt} (tot){_tm_finish - _tm_start} (finalize){_tm_finalize_trade - _tm_start} (calc){_tm_calc_trade - _tm_finalize_trade} (misc){_tm_finish - _tm_calc_trade}')

    def book_trades(self, trade_list: list):
        if trade_list and len(trade_list) > 0:
            self._pending_trades += trade_list

    def get_asset_result(self):
        return self._report_helper.get_asset_stat()

    def get_fund_result(self):
        return self._report_helper.get_fund_stat()

    def get_asset_trade(self):
        return self._report_helper.get_asset_trade()

    def get_fund_trade(self):
        return self._report_helper.get_fund_trade()
        
    def plot(self, is_asset:bool=True, is_fund:bool=True):
        if is_asset:
            self._report_helper.backtest_asset_plot()
        if is_fund:
            self._report_helper.backtest_fund_plot()
    
    def plot_score(self, index_id, is_tuning=False):
        #['csi500', 'gem', 'gold', 'hs300', 'national_debt', 'sp500rmb']
        self._report_helper._plot_fund_score(index_id, is_tuning)

    def plot_taa(self, saa_mv, taa_mv, index_id):
        #['csi500', 'hs300', 'gem', 'sp500rmb']
        self._report_helper._plot_taa_saa(saa_mv, taa_mv, index_id)
        self._report_helper._index_pct_plot(index_id, saa_mv, taa_mv)

    def get_last_position(self):
        return self._report_helper.get_last_position()

    def get_last_target_fund_allocation(self):
        return self._report_helper.get_last_target_fund_allocation(self._prep_target_fund_allocation.funds)

def saa_backtest(m: FundDataManager, saa: AssetWeight):
    asset_param = AssetTradeParam() # type in here
    t = AssetTrader(asset_param)
    b = FundBacktestEngine(data_manager=m, trader=t, taa_params=None)
    b.init()
    b.run(saa=saa)

def taa_backtest(m: FundDataManager, saa: AssetWeight):
    taa_param = TAAParam()  # type in here
    asset_param = AssetTradeParam() # type in here
    t = AssetTrader(asset_param)
    b = FundBacktestEngine(data_manager=m, trader=t, taa_params=taa_param)
    b.init()
    b.run(saa=saa)

def fund_backtest_without_taa(m: FundDataManager, saa: AssetWeight):
    asset_param = AssetTradeParam() # type in here
    fund_param = FundTradeParam() # type in here
    t = FundTrader(asset_param, fund_param)
    fa_param = FAParam() # type in here
    b = FundBacktestEngine(data_manager=m, trader=t, taa_params=None, fa_params=fa_param)
    b.init()
    b.run(saa=saa)

def fund_backtest(m: FundDataManager, saa: AssetWeight):
    asset_param = AssetTradeParam() # type in here
    fund_param = FundTradeParam(EnableCommission=True) # type in here
    t = FundTrader(asset_param, fund_param)
    taa_param = TAAParam()  # type in here
    fa_param = FAParam() # type in here
    b = FundBacktestEngine(data_manager=m, trader=t, taa_params=taa_param, fa_params=fa_param)
    b.init()
    b.run(saa=saa)

def fund_realtime(m: FundDataManager, saa: AssetWeight):
    asset_param = AssetTradeParam() # type in here
    fund_param = FundTradeParam() # type in here
    t = FundTrader(asset_param, fund_param)
    taa_param = TAAParam()  # type in here
    fa_param = FAParam() # type in here
    b = FundEngine(data_manager=m, trader=t, taa_params=taa_param, fa_params=fa_param)
    b.init()
    b.setup(saa)
    _dts = m.get_trading_days()
    dts = _dts[(_dts.datetime >= m.start_date) & (_dts.datetime <= m.end_date)].datetime
    calc_dt = dts.iloc[0]
    final_dt = dts.iloc[1]
    cash = b.DEFAULT_CASH_VALUE
    cur_asset_position = AssetPosition(cash=cash)
    cur_fund_position = FundPosition(cash=cash)

    b.prep_data(calc_dt)
    trade_list = b.calc_trade(calc_dt, cur_asset_position, cur_fund_position)
    b.prep_data(final_dt)
    traded_list = b.finalize_trade(final_dt, trade_list, cur_asset_position, cur_fund_position)
    print(trade_list)
    print(traded_list)
    print(cur_fund_position)
    print(cur_asset_position)

def taa_result(m: FundDataManager):
    start_date=datetime.date(2010,1,1)
    end_date=datetime.date(2020,5,20)
    index_id = 'hs300'
    taaParam = TAAParam(HighThreshold = 1,
                        HighStop = 0.45,
                        HighMinus = 0.07,
                        LowStop = 0.43,
                        LowThreshold = 0.15,
                        LowPlus = 0.06) 
    return ReportHelper.get_taa_result(index_id=index_id, start_date=start_date, end_date=end_date, taaParam=taaParam, dm=m)

def four_dimensions_backtest(m,begin_date, end_date, saa, fund_param, taa_detail):
    asset_param = AssetTradeParam() 
    t = AssetTrader(asset_param)
    saa_bk = FundBacktestEngine(data_manager=m, trader=t, taa_params=None)
    saa_bk.init()
    saa_bk.run(saa=saa,start_date=begin_date,end_date=end_date)
    saa_mv = saa_bk.get_asset_result()['market_value']
    asset_param = AssetTradeParam() 
    t = AssetTrader(asset_param)
    taa_bk = FundBacktestEngine(data_manager=m, trader=t, taa_params=None,taa_param_details=taa_detail)
    taa_bk.init()
    taa_bk.run(saa=saa,start_date=begin_date,end_date=end_date)
    taa_mv = taa_bk.get_asset_result()['market_value']
    asset_param = AssetTradeParam() 
    
    t = FundTrader(asset_param, fund_param)
    fa_param = FAParam() 
    fund_taa = FundBacktestEngine(data_manager=m, trader=t, taa_params=None, fa_params=fa_param, taa_param_details=taa_detail)
    fund_taa.init()
    fund_taa.run(saa=saa,start_date=begin_date,end_date=end_date)
    fund_result = fund_taa.get_fund_result()
    fund_mv = fund_result['market_value']
    fund_annual_ret = fund_result['annual_ret']
    fund_mv = fund_mv.rename(columns={'mv':'优选基金'})
    taa_mv = taa_mv.rename(columns={'mv':'战术配置'})
    saa_mv = saa_mv.rename(columns={'mv':'战略配置'})
    mv_df = pd.concat([fund_mv,taa_mv, saa_mv], axis =1)
    mv_df = mv_df.join(m.dts.index_price[['hs300']].rename(columns={'hs300':'沪深300'}))
    res = []
    for port_name in mv_df:
        res_i = Calculator.get_stat_result_from_df(df=mv_df.reset_index(), date_column='date', value_column=port_name).__dict__
        res_i['策略名称'] = port_name
        res.append(res_i)
    stat_df = pd.DataFrame(res).set_index('策略名称').reindex(['沪深300','战略配置','战术配置','优选基金'])[['annualized_ret','annualized_vol','sharpe','mdd','mdd_date1','mdd_date2']]
    return mv_df, stat_df, fund_annual_ret

def licaimofang_backtest(m: FundDataManager, saa:AssetWeight):
    json_path = '../../surfing/rpm/etc/licaimofang_white_list.json'
    with open(json_path,'r') as f:
        white_list = set(json.load(f))
    black_list = set(['000356!0','004870!0','008593!0','007404!0','003547!0'])  #基金代码在理财魔方白名单， 对方客服回复不能交易但基金代码

    taa_detail_item_mmf = TAAParam(HighThreshold = 0.95,
                                HighStop = 0.5,
                                HighMinus = 0.05,
                                LowStop = -1,
                                LowThreshold = -1,
                                LowPlus = 0,
                                ToMmf=True)

    taa_detail= {
        'hs300': taa_detail_item_mmf,
        'csi500': taa_detail_item_mmf,
        'gem': taa_detail_item_mmf,
        'sp500rmb': taa_detail_item_mmf,
    }
    asset_param = AssetTradeParam()

    fund_param = FundTradeParam(JudgeIndexDiff=0.06, #大类资产偏离度
                                JudgeFundSelection=0.4, # 基金排名偏离度
                                JudgeFundRebalance=0.8, # 基金比例平均化
                        )

    fa_param = FAParam(MaxFundNumUnderAsset=4 # 同一个大类资产下基金最多数
                    )
    taa_default = TAAParam()
    saa = AssetWeight(hs300=0.14,csi500=0.25,gem=0.20,sp500rmb=0.1,national_debt=0.3,cash=1/100)


    t = FundTrader(asset_param, fund_param)
    end_date = datetime.date.today()
    #end_date = datetime.date(2020,7,8)
    begin_date = datetime.date(2012,1,1)
    bk = FundBacktestEngine(data_manager=m, trader=t, taa_params=taa_default, fa_params=fa_param, taa_param_details=taa_detail,white_list=white_list,black_list=black_list)
    bk.init()
    bk.run(saa=saa, start_date=begin_date, end_date=end_date)
    bk_result = bk.get_fund_result()
    return bk_result

def qieman_backtest(m: FundDataManager, saa:AssetWeight):
    black_list_json = '../../surfing/rpm/etc/qieman_black_list.json'
    white_list_json = '../../surfing/rpm/etc/qieman_white_list.json'
    with open(white_list_json,'r') as f:
        white_list = set(json.load(f))        
    with open(black_list_json,'r') as f:
        black_list = set(json.load(f))
    taa_detail_item_mmf = TAAParam(HighThreshold = 0.95,
                                HighStop = 0.5,
                                HighMinus = 0.05,
                                LowStop = -1,
                                LowThreshold = -1,
                                LowPlus = 0,
                                ToMmf=True)

    taa_detail= {
        'hs300': taa_detail_item_mmf,
        'csi500': taa_detail_item_mmf,
        'gem': taa_detail_item_mmf,
        'sp500rmb': taa_detail_item_mmf,
    }
    asset_param = AssetTradeParam()

    fund_param = FundTradeParam(JudgeIndexDiff=0.06, #大类资产偏离度
                                JudgeFundSelection=0.4, # 基金排名偏离度
                                JudgeFundRebalance=0.8, # 基金比例平均化
                        )

    fa_param = FAParam(MaxFundNumUnderAsset=4 # 同一个大类资产下基金最多数
                    )
    taa_default = TAAParam()
    saa = AssetWeight(hs300=0.14,csi500=0.25,gem=0.20,sp500rmb=0.1,national_debt=0.3,cash=1/100)

    t = FundTrader(asset_param, fund_param)
    end_date = datetime.date.today()
    begin_date = datetime.date(2012,1,1)
    bk = FundBacktestEngine(data_manager=m, trader=t, taa_params=taa_default, fa_params=fa_param, taa_param_details=taa_detail,white_list=white_list,black_list=black_list)
    bk.init()
    bk.run(saa=saa, start_date=begin_date, end_date=end_date)
    bk_result = bk.get_fund_result()
    return bk_result

def change_score_function(m: FundDataManager, saa: AssetWeight):
    t = FundTrader(AssetTradeParam(), FundTradeParam())
    taa_param = TAAParam()
    fa_param = FAParam()
    funcs = {
        'hs300': ScoreFunc(alpha=0, beta=-1, fee_rate=0),
        'csi500': ScoreFunc(alpha=0, beta=-1)
    }
    bk = FundBacktestEngine(data_manager=m, trader=t, taa_params=taa_param, fa_params=fa_param, fund_score_funcs=funcs)
    bk.init()
    bk.run(saa=saa)
    bk_result = bk.get_fund_result()
    return bk_result

def change_formula_score_function(m: FundDataManager, saa: AssetWeight):
    t = FundTrader(AssetTradeParam(), FundTradeParam())
    taa_param = TAAParam()
    fa_param = FAParam()
    func_str = '0.5 * alpha + 0.1 * abs(1 - beta)'
    bk = FundBacktestEngine(data_manager=m, trader=t, taa_params=taa_param, fa_params=fa_param, fund_score_funcs={'hs300': func_str})
    bk.init()
    bk.run(saa=saa)
    bk_result = bk.get_fund_result()
    return bk_result

def test():
    from ...data.manager.score import FundScoreManager
    m = FundDataManager('20150101', '20160101', score_manager=FundScoreManager())
    m.init()

    saa = AssetWeight(
        hs300=15/100,
        csi500=5/100,
        gem=3/100,
        sp500rmb=7/100,
        national_debt=60/100,
        gold=10/100,
        cash=5/100
    )
    
    # saa_backtest(m, saa)
    # taa_backtest(m, saa)
    # fund_backtest_without_taa(m, saa)
    fund_backtest(m, saa)
    # fund_realtime(m, saa)
    # change_score_function(m, saa)
    # change_formula_score_function(m, saa)

if __name__ == '__main__':
    # profile(file_name='/Users/cjiang/taa_perf1.txt', func=fund_backtest)
    test()


import datetime
import pandas as pd
import numpy as np
import traceback
from ...api.raw import RawDataApi
from ...api.basic import BasicDataApi
from ...view.derived_models import IndexValuation
from .derived_data_helper import DerivedDataHelper


class IndexValProcess:

    YEAR = 242
    WINDOW = YEAR * 5
    MIN_PERIOD = YEAR * 3
    SCORE_METHOD = {
        'PB百分位': 'pb_pct',
        'PE百分位': 'pe_pct',
        'PS百分位': 'ps_pct',
    }
    TAA_INDEX_LIST = ['hs300', 'csi500', 'gem', 'sp500rmb']

    def __init__(self, data_helper):
        self._data_helper = data_helper
        self.raw_api = RawDataApi()
        self.basic_api = BasicDataApi()

    def init(self, start_time, end_time):
        self.start_time = datetime.datetime.strptime(start_time, '%Y%m%d').date()
        self.end_time = datetime.datetime.strptime(end_time, '%Y%m%d').date()
        self.index_info = self.basic_api.get_index_info()
        # 过滤掉em_id是null的情况
        self.index_info = self.index_info[self.index_info.em_id.notna()]
        em_id_to_index_id = self.index_info[['index_id', 'em_id']].set_index('em_id').to_dict()['index_id']
        self.index_val = self.raw_api.get_em_index_val(self.start_time, self.end_time, self.index_info.em_id.to_list())
        self.index_val['index_id'] = self.index_val['em_id'].map(lambda x: em_id_to_index_id[x])
        self.index_val = self.index_val.drop(columns=['em_id', '_update_time']).set_index(['index_id', 'datetime'])
        self.index_id_score_method_dict = self.index_info[self.index_info['tag_method'] != 'none'][['index_id', 'tag_method']].set_index('index_id').to_dict()['tag_method']

    def calculate(self):
        pctrank = lambda x: x.rank(pct=True).iloc[-1]
        self.index_val = self.index_val.rename(columns={'dividend_yield': 'dy'})
        index_list = self.index_val.index.levels[0].tolist()
        self.result = []
        for index_id in index_list:
            if index_id in self.TAA_INDEX_LIST:
                mins_periods = self.MIN_PERIOD
            else:
                mins_periods = 3 * self.YEAR
            df_i = self.index_val.loc[index_id].copy()
            df_i.loc[:, 'pe_pct'] = df_i['pe_ttm'].tail(self.WINDOW+1).rolling(window=self.WINDOW, min_periods=mins_periods).apply(pctrank)
            df_i.loc[:, 'pb_pct'] = df_i['pb_mrq'].tail(self.WINDOW+1).rolling(window=self.WINDOW, min_periods=mins_periods).apply(pctrank)
            df_i.loc[:, 'ps_pct'] = df_i['ps_ttm'].tail(self.WINDOW+1).rolling(window=self.WINDOW, min_periods=mins_periods).apply(pctrank)
            df_i.loc[:, 'roe_pct'] = df_i['roe'].tail(self.WINDOW+1).rolling(window=self.WINDOW, min_periods=mins_periods).apply(pctrank)

            df_i['peg_ttm'] = df_i['pe_ttm'] / (df_i['dy'] / df_i['dy'].shift(self.YEAR)-1)
            df_i.loc[:, 'index_id'] = index_id
            self.result.append(df_i)
            # print(f'finish {index_list.index(index_id)} total: {len(index_list)}')
        self.result = pd.concat(self.result).reset_index()
        score_list = []
        for r in self.result.itertuples(index=False):
            tag_method = self.index_id_score_method_dict.get(r.index_id, None)
            score_data = self.SCORE_METHOD.get(tag_method, 'pe_pct')
            score_list.append(getattr(r, score_data))

        self.result['datetime'] = pd.to_datetime(self.result['datetime'])
        self.result['val_score'] = score_list
        self.result = self.result.replace(np.inf, float('nan')).replace(-np.inf, float('nan'))

    def filter_est_peg(self, x):
        if x > 10000 or x < -10000:
            return None
        return x

    def process(self, start_date, end_date):
        failed_tasks = []
        try:
            start_date_dt = datetime.datetime.strptime(start_date, '%Y%m%d').date()
            end_date_dt = datetime.datetime.strptime(end_date, '%Y%m%d').date()
            start_date = start_date_dt - datetime.timedelta(days = 2000) #5年历史保险起见，多取几天 5*365=1825
            start_date = datetime.datetime.strftime(start_date, '%Y%m%d')
            self.init(start_date, end_date)
            self.calculate()
            self.result = self.result.dropna(subset=['pe_ttm','pb_mrq','ps_ttm'])
            self.result = self.result[self.result['datetime'] >= pd.to_datetime(start_date_dt)]
            self.result = self.result[self.result['datetime'] <= pd.to_datetime(end_date_dt)]
            self.result['est_peg'] = self.result['est_peg'].map(lambda x : self.filter_est_peg(x))
            self._data_helper._upload_derived(self.result, IndexValuation.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('index_val')

        return failed_tasks


if __name__ == "__main__":
    IndexValProcess(DerivedDataHelper()).process('20200727', '20200727')




import dataclasses
import datetime
import pandas
import numpy as np
import statsmodels.api as sm
from statsmodels import regression
from scipy.stats import gmean
from functools import partial
from ..data.api.basic import BasicDataApi
from ..data.struct import TAAParam, TaaTunerParam, AssetWeight
from ..data.api.derived import DerivedDataApi
from ..fund.engine.asset_helper import TAAHelper, TAAStatusMode
from ..data.wrapper.mysql import *
from ..data.view.basic_models import *
from ..data.view.derived_models import *

@dataclasses.dataclass
class SeriesStatUnit:

    start_date: datetime.date   # 序列的开始时间
    end_date: datetime.date     # 序列的结束时间
    trade_year: float           # 交易年限
    last_unit_nav: float        # 末期单位净值
    annualized_ret: float       # 年化收益
    annualized_vol: float       # 年化波动率
    sharpe: float               # 夏普率
    recent_y5_ret: float        # 近五年收益 
    mdd: float                  # 最大回撤
    mdd_date1: datetime.date    # 最大回撤开始时间
    mdd_date2: datetime.date    # 最大回撤结束时间
    ret_over_mdd: float         # Calmar 年化收益除以最大回撤
    last_mv_diff: float         # 末期市值增量              
    last_increase_rate: float   # 末期涨幅

@dataclasses.dataclass
class BenchmarkSeiesStatUnit(SeriesStatUnit):

    alpha: float                # 超额收益
    beta: float                 # 风险暴露
    track_err: float            # 跟踪误差
    ir: float                   # 信息比率

class Calculator:

    TRADING_DAYS_PER_YEAR = 242
    
    @staticmethod
    def get_risk_free_rate():
        return 0.025

    @staticmethod
    def get_stat_result(dates: pandas.core.series.Series,
                        values: pandas.core.series.Series):
        assert len(dates) == len(values), 'date and value has different numbers'
        
        if len(dates) > 1:
            sr = pandas.Series(values, index=dates).sort_index()
            risk_fee_rate = Calculator.get_risk_free_rate()
            start_date = sr.index[0] 
            end_date = sr.index[-1]
            trade_year = sr.shape[0] / Calculator.TRADING_DAYS_PER_YEAR
            last_unit_nav = sr[-1] / sr[0]
            annualized_ret = np.exp(np.log(last_unit_nav) / trade_year) - 1
            annualized_vol = (sr / sr.shift(1)).std(ddof=1) * np.sqrt(Calculator.TRADING_DAYS_PER_YEAR)
            sharpe = (annualized_ret - risk_fee_rate) / annualized_vol
            if sr.shape[0] > Calculator.TRADING_DAYS_PER_YEAR * 5:
                recent_y5_ret  = sr[-1] / sr[- Calculator.TRADING_DAYS_PER_YEAR * 5] - 1
            else:
                recent_y5_ret = sr[-1] / sr[0] - 1
            mdd_part =  sr[:] / sr[:].rolling(window=sr.shape[0], min_periods=1).max()
            mdd = 1 - mdd_part.min()
            mdd_date2 = mdd_part.idxmin()
            mdd_date1 = sr[:mdd_date2].idxmax()
            ret_over_mdd = annualized_ret / mdd if mdd != 0 else None
            last_mv_diff = sr[-1] - sr[-2]
            last_increase_rate = (sr[-1] - sr[-2])/ sr[-2]
        # 如果只输入一天，返回空
        else:
            start_date = dates[0]
            end_date = dates[0]
            trade_year = 0
            last_unit_nav = 1
            annualized_ret = 0
            annualized_vol = 0
            sharpe = 0 
            recent_y5_ret = 0
            mdd = 0
            mdd_date1 = dates[0]
            mdd_date2 = dates[0]
            ret_over_mdd = 0
            last_mv_diff = 0
            last_increase_rate = 0

        return SeriesStatUnit(
            start_date=start_date,
            end_date=end_date,
            trade_year=trade_year,
            last_unit_nav=last_unit_nav,
            annualized_ret=annualized_ret,
            annualized_vol=annualized_vol,
            sharpe=sharpe,
            recent_y5_ret=recent_y5_ret,
            mdd=mdd,
            mdd_date1=mdd_date1,
            mdd_date2=mdd_date2,
            ret_over_mdd=ret_over_mdd,
            last_mv_diff=last_mv_diff,
            last_increase_rate=last_increase_rate,
        )

    @staticmethod
    def get_benchmark_stat_result(dates: pandas.core.series.Series,
                                  values: pandas.core.series.Series,
                                  benchmark_values: pandas.core.series.Series):
        
        assert len(dates) == len(values), 'date and port_values has different numbers'
        assert len(dates) == len(benchmark_values), 'date and bench_values has different numbers'
        res = Calculator.get_stat_result(dates, values)
        sr_values  = pandas.Series(values, index=dates).sort_index()  
        sr_values = sr_values / sr_values[0]
        sr_benchmark_values = pandas.Series(benchmark_values, index=dates).sort_index()  
        sr_benchmark_values = sr_benchmark_values / sr_benchmark_values[0]
        sr_values_ret = sr_values / sr_values.shift(1)
        sr_benchmark_ret = sr_benchmark_values / sr_benchmark_values.shift(1)
        x = sr_benchmark_ret.dropna().values
        y = sr_values_ret.dropna().values
        x = sm.add_constant(x)
        model = regression.linear_model.OLS(y,x).fit()
        x = x[:,1]
        alpha = model.params[0]
        beta = model.params[1]
        track_err = (sr_values_ret - sr_benchmark_ret).std(ddof=1) * np.sqrt(Calculator.TRADING_DAYS_PER_YEAR)
        ir = alpha / track_err
        
        return BenchmarkSeiesStatUnit(
            start_date=res.start_date,
            end_date=res.end_date,
            trade_year=res.trade_year,
            last_unit_nav=res.last_unit_nav,
            annualized_ret=res.annualized_ret,
            annualized_vol=res.annualized_vol,
            sharpe=res.sharpe,
            recent_y5_ret=res.recent_y5_ret,
            mdd=res.mdd,
            mdd_date1=res.mdd_date1,
            mdd_date2=res.mdd_date2,
            ret_over_mdd=res.ret_over_mdd,
            alpha=alpha,
            beta=beta,
            track_err=track_err,
            ir=ir,
            last_mv_diff=res.last_mv_diff,
            last_increase_rate=res.last_increase_rate,
        )

    @staticmethod
    def get_benchmark(index_price: pandas.DataFrame,
                        weight_list: list,
                        index_list: list,
                        begin_date: datetime.date,
                        end_date: datetime.date)->pandas.DataFrame:

        assert len(weight_list) == len(index_list), 'weight_list and index_list has different numbers'
        assert index_price.shape[1] == len(weight_list), 'columns in index_price are not equal to index list'
        assert begin_date <= end_date, 'end date must bigger than begin date'
        if len(weight_list) < 1:
            return pandas.DataFrame()
        index_price = index_price.loc[begin_date:end_date,].fillna(method='bfill').fillna(method='ffill')
        index_price = index_price / index_price.iloc[0]
        weight_dict = [{'index_id':index_id, 'weight':weight_i } for index_id, weight_i in zip(index_list, weight_list)]
        weight_df = pandas.DataFrame(weight_dict)
        weight_df['weight'] = weight_df['weight'] / weight_df.weight.sum()
        weight_df = weight_df.set_index('index_id').T
        index_price['benchmark'] = index_price.mul(weight_df.values).sum(axis=1)
        return index_price[['benchmark']]

    @staticmethod
    def get_turnover_rate(dates: pandas.core.series.Series,
                          values: pandas.core.series.Series,
                          total_amount: float):
        year_num = len(list(set([_.year for _ in dates])))
        return total_amount / gmean(values) / year_num * 100

    @staticmethod
    def get_stat_result_from_df(df: pandas.DataFrame,
                                date_column: str,
                                value_column: str):
        dates = df[date_column].values
        values = df[value_column].values
        return Calculator.get_stat_result(dates, values)

    @staticmethod
    def get_benchmark_stat_result_from_df(df: pandas.DataFrame,
                                          date_column: str,
                                          port_value_column: str,
                                          bench_value_column: str):
        dates = df[date_column].values
        port_values = df[port_value_column].values
        bench_values = df[bench_value_column].values
        return Calculator.get_benchmark_stat_result(dates, port_values, bench_values)

    @staticmethod
    def get_benchmark_result(index_list:list,
                             weight_list:list,
                             begin_date: datetime.date,
                             end_date: datetime.date) -> pandas.DataFrame:
        '''
        index_list = ['hs300','national_debt']
        weight_list = [0.7,0.3]
        begin_date = datetime.date(2005,1,1)
        end_date = datetime.date(2020,1,1)
        Calculator.get_benchmark_result(index_list, weight_list, begin_date, end_date)
        '''                   
        basic = BasicDataApi()
        index_price = basic.get_index_price(index_list).pivot_table(index='datetime', columns='index_id', values='close')
        return Calculator.get_benchmark(index_price, weight_list, index_list, begin_date, end_date)
    
    @staticmethod
    def get_index_pct_data_df(index_id:str='hs300',
                           index_val:str=None,
                           windows:int=5*242,
                           min_periods:int=3*242,
                           pct_list:list=[0.15, 0.5, 0.6, 0.8]):
        
        assert (index_id in list(TaaTunerParam.PCT_PAIR.keys())) or (index_val is not None), \
            f'neither index_id{index_id} nor index_val {index_val} is valid'

        index_val = index_val if index_val else TaaTunerParam.PCT_PAIR[index_id]
        # load data
        index_val_data = DerivedDataApi().get_index_valuation_develop_without_date(index_ids=[index_id])
        
        values = index_val_data[index_val].values
        dates = index_val_data.datetime.values
        return  Calculator.get_index_pct_data(dates=dates,
                                            values=values,
                                            pct_list=pct_list,
                                            windows=windows,
                                            min_periods=min_periods)

    @staticmethod
    def roll_pct(x, pct_list, res):
        l = sorted(x)
        lens = len(l)
        res.append([l[int(lens*pct_i)] for pct_i in pct_list])
        return 1

    @staticmethod
    def get_index_pct_data(dates:pandas.core.series.Series,
                              values: pandas.core.series.Series,
                              pct_list:list,
                              windows:int,
                              min_periods:int):
        df = pandas.DataFrame(values, index=dates).dropna()
        res = []
        df[0].rolling(window= windows, min_periods=min_periods).apply(partial(Calculator.roll_pct, pct_list=pct_list, res=res), raw=True)
        return pandas.DataFrame(res, index=dates[-len(res):])

    @staticmethod
    def get_taa_detail(index_id:str='hs300',
                       index_val:str='roe_pct',
                       taa_param:TAAParam=TAAParam(),
                       ):
        
        # load data
        _dts = BasicDataApi().get_trading_day_list()
        dts = _dts.datetime 
        index_val_data = DerivedDataApi().get_index_valuation_develop_without_date(index_ids=[index_id])
        index_val_data = index_val_data.set_index('datetime').reindex(dts).fillna(method='ffill')
        asset_func_index = 'hs300'
        index_val_data['index_id'] = asset_func_index
        index_val_data = index_val_data.reset_index().set_index(['datetime','index_id'])
        fake_weight = AssetWeight(**{asset_func_index:1})
        taa_helper = TAAHelper(taa_params=taa_param)
        for dt in dts:
            asset_pct = index_val_data.loc[dt]
            taa_helper.on_price(dt=dt, asset_price=None, cur_saa=fake_weight, asset_pct=asset_pct, select_val={asset_func_index:index_val}, score_dict={})
        df = pandas.DataFrame(taa_helper.tactic_history).T.dropna()
        df['last_date'] = df.index.to_series().shift(1)
        con = df[asset_func_index] != df[asset_func_index].shift(1)
        df_diff_part = df[con].copy()
        df_diff_part = df_diff_part.reset_index().rename(columns={'index':'begin_date'})
        next_date = df_diff_part['last_date'].shift(-1).tolist()
        next_date[-1] = df.index.values[-1]
        df_diff_part['end_date'] = next_date
        df_result = df_diff_part[df_diff_part[asset_func_index] != TAAStatusMode.NORMAL][['begin_date','end_date',asset_func_index]]
        df_result = df_result.rename(columns = {asset_func_index:'status'})
        return df_result.to_dict('records') 


    @staticmethod
    def get_val_level(x):
        val_dict = {
            'PB百分位':'pb_pct',
            'PE百分位':'pe_pct',
            'PS百分位':'ps_pct',
        }
        pct = getattr(x, val_dict[x.tag_method])
        pe = x.pe_ttm
        if None in [pct,pe]:
            return None
        if pct < 0.3 and pe < 20:
            res = 'low'
        elif pct > 0.7 and pe > 20:
            res = 'high'
        else:
            res = 'median'
        return res

    @staticmethod
    def get_index_level(dt:datetime.date=datetime.date(2020,7,20)):
    
        with BasicDatabaseConnector().managed_session() as db_session:
            query = db_session.query(
                    IndexInfo
            )
            index_info = pandas.read_sql(query.statement, query.session.bind)
        with DerivedDatabaseConnector().managed_session() as mn_session:
            query = mn_session.query(
                IndexValuation
            ).filter(
                IndexValuation.datetime== dt,
            )
            index_val_df = pandas.read_sql(query.statement, query.session.bind)
        index_val = index_val_df.set_index('index_id').join(index_info.set_index('index_id')[['tag_method']]).reset_index()
        index_val['val_level'] = index_val.apply(lambda x : Calculator.get_val_level(x), axis=1 )
        return index_val[['index_id','val_level']]

    @staticmethod
    def get_index_level_index(index_id:str='hs300'):
        with BasicDatabaseConnector().managed_session() as db_session:
            query = db_session.query(
                    IndexInfo
            )
            index_info = pandas.read_sql(query.statement, query.session.bind)
        with DerivedDatabaseConnector().managed_session() as mn_session:
            query = mn_session.query(
                IndexValuation
            ).filter(
                IndexValuation.index_id.in_([index_id]),
            )
            index_val_df = pandas.read_sql(query.statement, query.session.bind)
        index_val = index_val_df.set_index('index_id').join(index_info.set_index('index_id')[['tag_method']]).reset_index()
        index_val['val_level'] = index_val.apply(lambda x : Calculator.get_val_level(x), axis=1 )
        index_val = index_val[['datetime','val_level']]
        res = []
        last_level = 'begin'
        dic = {}
        for i in index_val.itertuples():
            if i.val_level != last_level and not dic:
                dic.update({'start_date':i.datetime})
                dic.update({'val_level':i.val_level})
                last_level = i.val_level

            elif i.val_level != last_level and dic:
                dic.update({'end_date':last_date})
                last_level = i.val_level
                res.append(dic)
                dic = {}
                dic.update({'start_date':i.datetime})
                dic.update({'val_level':i.val_level})
            last_date = i.datetime
        if 'end_date' not in dic:
            dic.update({'end_date':last_date}) 
        res.append(dic)
        res = [i for i in res if i['val_level'] != 'median']
        return res

    @staticmethod
    def get_mdd_recover_result(mv_df:pandas.DataFrame):

        def _get_mdd_recover_result_df(df:pandas.DataFrame, date_column:str, value_column:str):
            dates = df[date_column].values
            values = df[value_column].values
            return _mdd_recover_analysis(values,dates)

        def _mdd_recover_analysis(values,dates):
            sr = pandas.Series(values, index=dates).sort_index()
            if sr.empty:
                mdd = 0
                mdd_date1 = None
                mdd_date2 = None
                mdd_lens = 0
                return mdd, mdd_date1, mdd_date2, mdd_lens
            mdd_part =  sr[:] / sr[:].rolling(window=sr.shape[0], min_periods=1).max()
            mdd = 1 - mdd_part.min()
            if mdd == 0:
                mdd_date1 = None
                mdd_date2 = None
                mdd_lens = 0
            else:
                mdd_date = mdd_part.idxmin()
                mdd_date1 = sr[:mdd_date].idxmax()
                sr_tmp = sr[mdd_date1:]
                recover_sr = sr_tmp[sr_tmp> sr[mdd_date1]]
                if recover_sr.empty:
                    mdd_date2 = sr_tmp.index[-1]
                else: 
                    mdd_date2 = sr_tmp[sr_tmp> sr[mdd_date1]].index[0]
                mdd_lens = sr.loc[mdd_date1:mdd_date2].shape[0]
            return mdd, mdd_date1, mdd_date2, mdd_lens

        mdd1, mdd_date1_1, mdd_date1_2, mdd_lens1 = _get_mdd_recover_result_df(mv_df.reset_index(),'date','mv')
        mv_df_1 = mv_df.loc[:mdd_date1_1].reset_index()
        mv_df_2 = mv_df.loc[mdd_date1_2:].reset_index()

        mdd2, mdd_date2_1, mdd_date2_2, mdd_lens2 = _get_mdd_recover_result_df(mv_df_1,'date','mv')
        mdd3, mdd_date3_1, mdd_date3_2, mdd_lens3 = _get_mdd_recover_result_df(mv_df_2,'date','mv')
        if mdd3 > mdd2:
            mdd2 = mdd3
            mdd_date2_1 = mdd_date3_1
            mdd_date2_2 = mdd_date3_2
            mdd_lens2 = mdd_lens3

        res = {
            'mdd1':mdd1,
            'mdd1_d1':mdd_date1_1,
            'mdd1_d2':mdd_date1_2,
            'mdd1_lens':mdd_lens1,
            'mdd2':mdd2,
            'mdd2_d1':mdd_date2_1,
            'mdd2_d2':mdd_date2_2,
            'mdd2_len':mdd_lens2
        }

        return res



if __name__ == "__main__":
    # get pct with select rank
    Calculator.get_index_pct_data_df()
    Calculator.get_index_pct_data_df(index_id='hs300',
                           index_val='roe',
                           windows=5*242,
                           min_periods=3*242,
                           pct_list=[0.15, 0.5, 0.6, 0.8])

    # get taa detail
    Calculator.get_taa_detail()
    Calculator.get_taa_detail(index_id='hs300',
                              index_val='pb_pct',
                              taa_param=TAAParam(),
                              )

    Calculator.get_index_level_index(index_id='hs300')


import dataclasses
import copy
import datetime
import numpy as np

@dataclasses.dataclass
class AssetDict:
    # 战略配置参数
    hs300: float=0.0
    csi500: float=0.0
    gem: float=0.0
    sp500rmb: float=0.0
    national_debt: float=0.0
    gold: float=0.0
    cash: float=0.0
    mmf: float=0.0
    hsi: float=0.0
    def __repr__(self):
        s = f'<{self.__class__.__name__}'
        for k, v in self.__dict__.items():
            s += f' {k}={v}'
        s += '>'
        return s

    def copy(self):
        return copy.deepcopy(self)

    def isnan(self, v):
        return v is None or np.isnan(v)

@dataclasses.dataclass
class AssetTimeSpan:
    # 大类资产指标考察年数
    hs300: int=3
    csi500: int=3
    gem: int=1
    sp500rmb: int=1
    national_debt: int=1
    gold: int=1
    mmf: int=1
    hsi: int=1

@dataclasses.dataclass
class AssetWeight(AssetDict):

    def __post_init__(self):
        self.rebalance()

    def rebalance(self):
        _sum = 0.0
        for k, v in self.__dict__.items():
            assert round(v,6) >= 0, f'{k} {v} has negative value'
            _sum += v
        assert _sum > 0, 'param sum value must be positive'
        # normalize, make sure change is correct
        for k, v in self.__dict__.items():
            self.__dict__[k] /= _sum
    
    def __repr__(self):
        s = f'<{self.__class__.__name__}'
        for k, v in self.__dict__.items():
            s += f' {k}={v*100}%'
        s += '>'
        return s

@dataclasses.dataclass
class AssetPrice(AssetDict):

    def __post_init__(self):
        for k, v in self.__dict__.items():
            assert v is None or np.isnan(v) or round(v,8) >= 0, f'{k} has negative value: {v}'
        self.cash = 1 # cash price is always 1
    
    def __repr__(self):
        s = f'<{self.__class__.__name__}'
        for k, v in self.__dict__.items():
            s += f' {k}={v}'
        s += '>'
        return s

@dataclasses.dataclass
class AssetTrade:

    index_id: str
    is_buy: bool
    mark_price: float
    submit_date: datetime.date
    amount: float=None
    volume: float=None
    trade_price: float=None
    commission: float=0.0
    trade_date: datetime.date=None

    def __post_init__(self):
        assert self.amount or self.volume, \
            f'both amount and volume are not valid (index){self.index_id} (vol){self.volume} (amt){self.amount}'

@dataclasses.dataclass
class AssetPosition(AssetDict):

    def __post_init__(self):
        for k, v in self.__dict__.items():
            assert round(v,8) >= 0, f'asset backtest {k} has negative value: {v}'
    
    def __repr__(self):
        s = f'<{self.__class__.__name__}'
        for k, v in self.__dict__.items():
            s += f' {k}={v}'
        s += '>'
        return s

    def update(self, trd: AssetTrade):
        p = trd.trade_price if trd.trade_price else trd.mark_price
        v = trd.volume if trd.volume else trd.amount / p
        assert trd.is_buy or (self.__dict__[trd.index_id] - v > -1e-8), f'asset trade failed, (sell){v} (cur){self.__dict__[trd.index_id]} (trd){trd} (pos){self}'
        amt = trd.amount if trd.amount else v * p
        self.__dict__[trd.index_id] += v if trd.is_buy else -v
        self.cash += (-amt if trd.is_buy else amt)

@dataclasses.dataclass
class AssetValue(AssetDict):

    prices: AssetPrice=None
    positions: AssetPosition=None

    def __post_init__(self):
        for index_id, p in self.prices.__dict__.items():
            self.__dict__[index_id] = 0 if self.isnan(p) else p * self.positions.__dict__[index_id]
            # 不允许出现价格为空，但是仓位为正的情况
            assert not (self.isnan(p) and self.positions.__dict__[index_id] > 0), \
                f'price is nan but position is not zero... (index){index_id} (p){p} (pos){self.positions.__dict__[index_id]}'

    def sum(self):
        _sum = 0.0
        for k in AssetDict.__dataclass_fields__.keys():
            _sum += self.__dict__[k]
        return _sum

    def __repr__(self):
        s = f'<{self.__class__.__name__}'
        for k, v in self.__dict__.items():
            s += f' {k}={v}'
        s += '>'
        return s

    def get_weight(self):
        wgts = AssetWeight(cash=1)
        for index_id, p in self.prices.__dict__.items():
            wgts.__dict__[index_id] = self.__dict__[index_id]
        wgts.rebalance()
        return wgts
        
@dataclasses.dataclass
class FundTrade(AssetTrade):

    fund_id: str=None
    fund_unit_nav: float=None
    fund_unit_volume: float=None
    is_to_cleanup: bool=False
    is_permitted_fund: bool=None
    trigger: int=None # by TradeTrigger in constant.py

@dataclasses.dataclass
class FundPosItem:

    fund_id: str            # 基金ID
    index_id: str           # 基金所属的大类资产
    volume: float             # 基金份额
    avg_cost: float=None    # 平均成本
    update_time: datetime.date=None # 更新时间
    is_permitted_fund: bool=None  # 在白名单里或者没有白名单 True   不在白名单里或者在黑名单里  False    

    def update(self, trade: FundTrade):
        assert self.fund_id == trade.fund_id, f'fund trade failed, self.fund_id != trade.fund_id: self.fund_id {self.fund_id}  trade.fund_id {trade.fund_id}'
        p = trade.trade_price if trade.trade_price else trade.mark_price
        v = trade.volume if trade.volume else trade.amount / p
        a = trade.amount if trade.amount else v * p
        assert trade.is_buy or trade.is_to_cleanup or (self.volume - v > -1e-8), f'fund trade failed, trade.volume {v} more than self.volume {self.volume} on {self.fund_id}'

        vol_diff = v if trade.is_buy else -v
        amt_diff = a if trade.is_buy else -a
        volume = self.volume + vol_diff
        if trade.is_buy:
            amt = self.volume * self.avg_cost + amt_diff# 买入时平均成本摊销
            self.volume = volume
            self.avg_cost = amt / volume
        else:
            self.volume -= trade.volume
            if self.volume < 1e-8:# 清空仓位 平均成本为0
                self.avg_cost = 0
            else: #非清仓卖出 平均成本不变
                pass
        self.update_time = trade.trade_date
        assert self.volume >= -1e-8, f'fund backtest vol {self.volume} has negative value: '
        assert self.avg_cost >= -1e-8, f"fund avg cost can not be negative! (fund_id){self.fund_id} (update time){self.update_time}"
        return True
        
    def __repr__(self):
        return f'<{self.fund_id} (v){self.volume} (c){self.avg_cost}>'

@dataclasses.dataclass
class AssetTradeRateItem(AssetDict):
    def __post_init__(self):
        for k, v in self.__dict__.items():
            self.__dict__[k] = None if (v == 0 and k != 'mmf') else v

@dataclasses.dataclass
class AssetTradeParam:

    MinCountAmtDiff: float=0.001 # 少于总资产千分之一的资产Amount差异，不计在内
    MinActionAmtDiff: float=0.06 # 如果没有资产的偏离达到这个比例，那就不调仓
    EnableCommission: bool=True # 是否开启交易手续费计算
    PurchaseDiscount: float=0.15  # 大类资产的买入折扣
    RedeemDiscount: float=1.0    # 大类资产的卖出折扣
    AssetPurchaseRate: dict=None 
    AssetRedeemRate: dict=None

    def __post_init__(self):
        self.AssetPurchaseRate = self.AssetPurchaseRate or AssetTradeRateItem(csi500=0.011055, 
                                                                                gem=0.010488, 
                                                                                gold=0.001892, 
                                                                                hs300=0.011987, 
                                                                                national_debt=0.005438, 
                                                                                sp500rmb=0.008135,
                                                                                mmf=0).__dict__ #大类资产平均买入费率
        self.AssetRedeemRate = self.AssetRedeemRate or AssetTradeRateItem(csi500=0.002973, 
                                                                            gem=0.003407,
                                                                            gold=0.000755,
                                                                            hs300=0.004727,
                                                                            national_debt=0.000287,
                                                                            sp500rmb=0.003240,
                                                                            mmf=0).__dict__ #大类资产平均卖出费率
@dataclasses.dataclass
class FundTradeParam:

    MaxFundNumUnderAsset: float=3 # 同一个大类资产下基金最多数
    MinCountAmtDiff: float=0.0003 # 少于总资产万分之三的资产Amount差异，不做基金调仓处理
    MinActionAmtDiff: float=0.03 # 如果没有资产的基金市值偏离达到这个比例，那就不调仓
    EnableCommission: bool=True # 是否开启交易手续费计算
    PurchaseDiscount: float=0.15 # 基金的买入折扣
    RedeemDiscount: float=1.0    # 基金的卖出折扣

    # new trading opt
    JudgeIndexDiff: float=0.06
    JudgeFundSelection: float=0.4 # 0-1，越大表示对基金选择的分数要求越高
    JudgeFundRebalance: float=0.8 # 0-1，越大表示对基金比例分配分散度要求越高
    DiffJudgeAssetWgtRequirement: float=0.01 # 比例超过该值的资产内的基金异常可以触发调仓

@dataclasses.dataclass
class TAAParam:

    HighThreshold: float=0.95    # 超过多少算作进入估值偏高的区域
    LowThreshold: float=-1     # 低于多少算作进入估值偏低的区域
    HighMinus:    float=0.05    # 估值偏高时应该调低的占比
    LowPlus:      float=0   # 估值偏低是应该调高的占比
    HighStop:     float=0.5     # 进入估值偏高区后，什么时候算作退出
    LowStop:      float=-1     # 进入估值偏低区后，什么时候算作退出
    TuneCash:     bool=False    # 默认taa调仓时，不调整cash
    ToMmf:        bool=True    # 默认taa加减仓不改变货基权重

@dataclasses.dataclass
class FundScoreParam:
    tag_type: int=1
    score_method: int=1
    is_full: int=1
    IndicatorTimespan: int=3    # 计算因子参数的时间要求

    def __repr__(self):
        return f'<fund_score_param tag_type={self.tag_type} score_method={self.score_method} is_full={self.is_full}'


@dataclasses.dataclass
class FAParam:

    MaxFundNumUnderAsset: float=4 # 同一个大类资产下基金最多数

@dataclasses.dataclass
class AssetTradeCache:
    
    index_id: str
    index_tar_wgt: float=None
    index_cur_wgt: float=None
    cur_fund_ids: set=None
    fund_scores: list=None
    fund_ranks: dict=None
    proper_fund_num: float=None
    fund_cur_wgts: dict=None
    fund_judge_ranking_score: float=None # 当前“基金选择”分数
    fund_judge_ranking_best: float=None  # “基金选择”最优分数
    fund_judge_ranking: float=None       # 0-1 之间，衡量基金选择分数
    fund_judge_diverse: float=None       # 0-1 之间，衡量基金配置比例

@dataclasses.dataclass
class FundWeightItem:

    fund_id: str                # 基金ID
    index_id: str               # 基金所属的大类资产
    asset_wgt: float            # 资产的wgt
    fund_wgt_in_asset: float    # 基金在所属资产中的wgt
    fund_wgt: float=None        # 基金在整体组合中的wgt

    def __post_init__(self):
        self.fund_wgt = self.fund_wgt or self.asset_wgt * self.fund_wgt_in_asset

    def add_tot_wgt(self, tot_wgt_to_add):
        self.fund_wgt += tot_wgt_to_add
        self.fund_wgt_in_asset = self.fund_wgt / self.asset_wgt

    def __repr__(self):
        return f'<{self.fund_id} (wgt){self.fund_wgt} (index){self.index_id} (iwgt){self.asset_wgt} (fwgt){self.fund_wgt_in_asset}>'

class FundWeight:

    def __init__(self):
        self.funds = {} # fund_id -> FundWeightItem
        self.index_fund_list = {} # index_id -> [fund_id]

    def add(self, fund_wgt_item: FundWeightItem):
        index_id = fund_wgt_item.index_id
        self.funds[fund_wgt_item.fund_id] = fund_wgt_item
        if index_id not in self.index_fund_list:
            self.index_fund_list[index_id] = []
        self.index_fund_list[index_id].append(fund_wgt_item.fund_id)
    
    def get_wgt(self, fund_id: str):
        fund_wgt_item = self.funds.get(fund_id)
        return fund_wgt_item.fund_wgt if fund_wgt_item else 0
    
    def get_funds(self, index_id):
        return self.index_fund_list.get(index_id, [])

    def get(self, fund_id: str):
        return self.funds.get(fund_id)

    def __repr__(self):
        s = f'<FW'
        for fund_id, wgt_item in self.funds.items():
            s += f' {wgt_item}'
        s += '>'
        return s

class FundPosition:

    def __init__(self, cash, funds=None, index_fund_list=None):
        self.funds = funds or {} # fund_id -> FundPosItem
        self.index_fund_list = index_fund_list or {} # index_id -> [fund_id]
        self.cash = cash

    def init_by_item_dict(self, cash, funds, index_fund_list):
        self.cash = cash
        self.funds = funds
        self.index_fund_list = index_fund_list

    def convert_to_item_dict(self):
        return {
            'cash': self.cash,
            'funds': {f: v.__dict__ for f, v in self.funds.items()},
            'index_fund_list': self.index_fund_list,
        }

    def calc_mv_n_w(self, 
                    fund_navs: dict, # fund_id -> nav(adjusted net value)
                    index_id=None): # filter: 只计算某个index_id
        mv = 0
        w_dic = {}
        for pos_item in self.funds.values():
            if index_id and pos_item.index_id != index_id or pos_item.volume == 0:
                continue
            nav = fund_navs.get(pos_item.fund_id, np.nan)
            if np.isnan(nav):
                print('warning!!!! nav in nan')
            else:
                mv += nav * pos_item.volume
            w_dic[pos_item.fund_id] = nav * pos_item.volume
            
        if index_id is None:
            mv = self.cash + mv
        w_dic = { f: w / mv for f, w in w_dic.items() }
        return mv, w_dic

    def calc_asset_position(self,
            fund_navs: dict,  # fund_id -> nav(adjusted net value)
            current_index_prices: AssetPrice,  # index_id -> nav(adjusted net value)
        ):
        mv_dic = {}
        for pos_item in self.funds.values():
            nav = fund_navs.get(pos_item.fund_id, np.nan)
            assert not np.isnan(nav), f'fund nav vol {pos_item.fund_id} has negative nav!'
            mv_dic[pos_item.index_id] = nav * pos_item.volume + mv_dic.get(pos_item.index_id, 0)
        vol_dic = {f: m / current_index_prices.__dict__.get(f) for f, m in mv_dic.items()}
        vol_dic['cash'] = self.cash
        return AssetPosition(**vol_dic)

    def get_volume(self, fund_id):
        item = self.funds.get(fund_id)
        if not item:
            return None
        return item.volume

    def get_funds(self, index_id):
        return self.index_fund_list.get(index_id, [])

    def add(self, fund_id, index_id, volume, avg_cost, update_time, is_permitted_fund):
        pos_item = FundPosItem(fund_id=fund_id, index_id=index_id, volume=volume, avg_cost=avg_cost, update_time=update_time, is_permitted_fund=is_permitted_fund)
        self.funds[fund_id] = pos_item
        if index_id not in self.index_fund_list:
            self.index_fund_list[index_id] = []
        self.index_fund_list[index_id].append(fund_id)
        return pos_item

    def update(self, trade: FundTrade):
        assert isinstance(trade, FundTrade), 'only FundTrade is accepted here'
        pos_item = self.funds.get(trade.fund_id)
        if pos_item is None:
            assert trade.is_buy, "fund volume cannot be negative"
            pos_item = self.add(trade.fund_id, trade.index_id, 0, 0, trade.trade_date or trade.submit_date, trade.is_permitted_fund)

        p = trade.trade_date if trade.trade_price else trade.mark_price
        amt = trade.amount if trade.amount else trade.volume * p
        cash_left = self.cash + (-amt if trade.is_buy else amt)
        if cash_left <= -1e-6:
            print(f'cash negative alert: cancel order due to cash negative trade: {trade}, cash left if trade successed: {cash_left}')
            return False
        if pos_item.update(trade):
            self.cash = cash_left
            return True
        else:
            return False

    def copy(self):
        return FundPosition(
            cash=self.cash, 
            funds=copy.deepcopy(self.funds), 
            index_fund_list=copy.deepcopy(self.index_fund_list)
        )

    '''
    # depreciated
    def calc_weight(self, 
                fund_navs: dict, # fund_id -> nav(adjusted net value)
                index_id=None): # filter: 只计算某个index_id
        mv = 0
        for pos_item in self.funds.values():
            if index_id and pos_item.index_id != index_id:
                continue
            nav = fund_navs.get(pos_item.fund_id, np.nan)
            if np.isnan(nav):
                print('warning!!!! nav in nan')
            else:
                mv += nav * pos_item.volume
        return mv
    '''

    def __repr__(self):
        s = f'<FP cash={self.cash}'
        for fund_id, pos_item in self.funds.items():
            s += f' {pos_item}'
        s += '>'
        return s

class FundScoreHelper:
    ASSET_DICT = {
        '利率债' : 'national_debt',
        '美股大盘': 'sp500rmb',
        '德国大盘': 'dax30rmb',
        '信用债': 'credit_debt',
        '房地产': 'real_state',
        'A股大盘': 'hs300',
        '原油': 'oil',
        '黄金': 'gold',
        '创业板': 'gem',
        '日本大盘': 'n225rmb',
        'A股中盘': 'csi500',
    }
    TACTIC_ASSET_CONVERT = {
        '000300.XSHG': 'hs300',
        '000905.XSHG': 'csi500'
    }

    @staticmethod
    def keys():
        return FundScoreHelper.ASSET_DICT.keys()

    @staticmethod
    def get(index_id):
        for k, v in FundScoreHelper.ASSET_DICT.items():
            if v == index_id:
                return k
        return None
    
    @staticmethod
    def parse(tag_name):
        return FundScoreHelper.ASSET_DICT.get(tag_name)

class TaaTunerParam:
    index_id:            str
    IsTaaLowOnly:        bool=False
    IsTaaUpOnly:         bool=False
    param_list: list=[] # 调参范围列表
    POOL = {
        'gem'   : 'ps_pct',
        'hs300' : 'pe_pct',
        'csi500': 'pe_pct',
        'sp500rmb': 'pe_pct', 
    }
    PCT_PAIR = {
        'gem':'ps_ttm',
        'hs300':'pe_ttm',
        'csi500':'pe_ttm',
        'sp500rmb':'pe_ttm',
    }

@dataclasses.dataclass
class ScoreFilter:

    FilterIndexId:list = None
    CompanyHoldLimit:int = 95
    SizeBottom:int = 1e8
    SizeTop:int = 1e10
    
    def __post_init__(self):
        self.FilterIndexId = self.FilterIndexId or ['national_debt']

@dataclasses.dataclass
class ScorePenaltyParams:

    FilterYearIndex:list = None
    Penalty:float = -0.5 
    JudgeYearLength:float = 1.0
    BlackListPenalty:float = - 10000

    def __post_init__(self):
        self.FilterYearIndex = self.FilterYearIndex or ['national_debt', 'csi500']

if __name__ == '__main__':
    a = AssetWeight(hs300=0.1, csi500=0.1, cash=0.8)
    print(a)


import pandas as pd
import numpy as np
import datetime
import json
import traceback
from ...wrapper.mysql import BasicDatabaseConnector, DerivedDatabaseConnector
from ...api.basic import *
from ...view.derived_models import *

class IndexIndicatorProcessor(object):
    w1 = 5
    m1 = 20
    m3 = 60
    m6 = 121
    y1 = 242
    y3 = 242*2
    y5 = 242*5
    y10 = 242*10

    def __init__(self, data_helper):
        self._data_helper = data_helper
        
    def history_init(self):
        with BasicDatabaseConnector().managed_session() as db_session:    
            query = db_session.query(IndexPrice).filter(IndexPrice.datetime >= datetime.date(2010,1,1))
            self.index_price = pd.read_sql(query.statement, query.session.bind)
            self.index_price = self.index_price.pivot_table(index='datetime', columns='index_id', values='close').fillna(method='ffill').fillna(method='bfill')
            self.date_list = self.index_price.index

    def init(self, end_date:str='20200715'):
        self.end_date = datetime.datetime.strptime(end_date,'%Y%m%d').date()
        with BasicDatabaseConnector().managed_session() as db_session:    
            query = db_session.query(IndexPrice).filter(IndexPrice.datetime <= self.end_date,
                                                        IndexPrice.datetime >= datetime.date(2005,1,1))
            self.index_price = pd.read_sql(query.statement, query.session.bind)
            self.index_price_ret = self.index_price.pivot_table(index='datetime', columns='index_id', values='close').fillna(method='ffill').fillna(method='bfill')
            self.index_price_vol_part = self.index_price.pivot_table(index='datetime', columns='index_id', values='close')
            query = db_session.query(TradingDayList)
            trade_day_df = pd.read_sql(query.statement, query.session.bind).drop(columns='_update_time')
            begin_d = self.index_price_vol_part.index[0]
            end_d = self.index_price_vol_part.index[-1]  
            trade_day_list = trade_day_df.datetime
            trade_day_index = trade_day_list[(trade_day_list>=begin_d) & (trade_day_list<=end_d)]
            self.index_price_vol_part = self.index_price_vol_part.reindex(trade_day_index).pct_change(1)
            self.date_list = self.index_price_ret.index

    def _get_begin_of_year(self, dt):
        _dt = datetime.date(dt.year,1,1)
        return self.date_list[self.date_list >= _dt][0]

    def _get_begin_of_season(self, dt):
        _m = int((dt.month-1) / 3) * 3 + 1
        _dt = datetime.date(dt.year, _m, 1)
        return self.date_list[self.date_list >= _dt][0]

    def _term_ret(self, df):
        return df.iloc[-1] / df.iloc[0] - 1

    def _term_vol(self, df):
        return df.std(ddof=1)

    def _this_y_ret(self, x):
        dt2 = x.name
        dt1 = self._get_begin_of_year(dt2)
        return self._term_ret(self.index_price.loc[dt1:dt2])

    def _this_s_ret(self, x):
        dt2 = x.name
        dt1 = self._get_begin_of_season(dt2)
        return self._term_ret(self.index_price.loc[dt1:dt2])

    def _cumulative_ret(self, x):
        dt = x.name
        return self._term_ret(self.index_price.loc[:dt])

    def _this_y_vol(self, x):
        dt2 = x.name
        dt1 = self._get_begin_of_year(dt2)
        return self._term_vol(self.index_price.loc[dt1:dt2])

    def _this_s_vol(self, x):
        dt2 = x.name
        dt1 = self._get_begin_of_season(dt2)
        return self._term_vol(self.index_price.loc[dt1:dt2])

    def _cumulative_vol(self, x):
        dt = x.name
        return self._term_vol(self.index_price.loc[:dt])

    def _ret_history(self):
        self.ret_res = {}
        self.ret_res['w1_ret'] = self.index_price.pct_change(self.w1)
        self.ret_res['m1_ret'] = self.index_price.pct_change(self.m1)
        self.ret_res['m3_ret'] = self.index_price.pct_change(self.m3)
        self.ret_res['m6_ret'] = self.index_price.pct_change(self.m6)
        self.ret_res['y1_ret'] = self.index_price.pct_change(self.y1)
        self.ret_res['y3_ret'] = self.index_price.pct_change(self.y3)
        self.ret_res['y5_ret'] = self.index_price.pct_change(self.y5)
        self.ret_res['y10_ret'] = self.index_price.pct_change(self.y10)
        self.ret_res['this_s_ret'] = self.index_price.apply(self._this_s_ret, axis=1)
        self.ret_res['this_y_ret'] = self.index_price.apply(self._this_y_ret, axis=1)
        self.ret_res['cumulative_ret'] = self.index_price.apply(self._cumulative_ret, axis=1)

    def _process_result(self, result):
        _res = []
        for k, df in result.items():
            _res.append(pd.DataFrame(df.stack()).rename(columns={0:k}))
        self.result = pd.concat(_res,axis=1,sort=False).reset_index()

    def process_ret_history(self):
        self._ret_history()
        self._process_result(self.ret_res)
        df = self.result[self.result.datetime > datetime.date(2018,1,1)]
        df = df.replace(np.Inf,None).replace(-np.Inf,None)
        df = df.drop_duplicates(subset=['index_id','datetime'])
        self._data_helper._upload_derived(df, IndexReturn.__table__.name)

    def _vol_history(self):
        self.ret_vol = {}
        self.ret_vol['w1_vol'] = self.index_price.rolling(self.w1).std(ddof=1)
        self.ret_vol['m1_vol'] = self.index_price.rolling(self.m1).std(ddof=1)
        self.ret_vol['m3_vol'] = self.index_price.rolling(self.m3).std(ddof=1)
        self.ret_vol['m6_vol'] = self.index_price.rolling(self.m6).std(ddof=1)
        self.ret_vol['y1_vol'] = self.index_price.rolling(self.y1).std(ddof=1)
        self.ret_vol['y3_vol'] = self.index_price.rolling(self.y3).std(ddof=1)
        self.ret_vol['y5_vol'] = self.index_price.rolling(self.y5).std(ddof=1)
        self.ret_vol['y10_vol'] = self.index_price.rolling(self.y10).std(ddof=1)
        self.ret_vol['this_s_vol'] = self.index_price.apply(self._this_s_vol, axis=1)
        self.ret_vol['this_y_vol'] = self.index_price.apply(self._this_y_vol, axis=1)
        self.ret_vol['cumulative_vol'] = self.index_price.apply(self._cumulative_vol, axis=1)

    def process_vol_history(self):
        self._vol_history()
        self._process_result(self.ret_vol)
        df = self.result[self.result.datetime > datetime.date(2018,1,1)]
        df = df.replace(np.Inf,None).replace(-np.Inf,None)
        df = df.drop_duplicates(subset=['index_id','datetime'])
        self._data_helper._upload_derived(df, IndexVolatility.__table__.name)

    def _process_update_result(self, result):
        res = []
        for k, s in result.items():
            s.name = k
            res.append(pd.DataFrame(s).rename(columns={0:k}))
        result = pd.concat(res, sort=False, axis=1).reset_index()
        result['datetime'] = self.end_date
        return result

    def _update_ret(self):
        y_date = self._get_begin_of_year(self.end_date)
        s_date = self._get_begin_of_season(self.end_date)
        self.ret_res = {}
        self.ret_res['w1_ret'] = self.index_price_ret.tail(self.w1+1).pct_change(self.w1).iloc[-1]
        self.ret_res['m1_ret'] = self.index_price_ret.tail(self.m1+1).pct_change(self.m1).iloc[-1]
        self.ret_res['m3_ret'] = self.index_price_ret.tail(self.m3+1).pct_change(self.m3).iloc[-1]
        self.ret_res['m6_ret'] = self.index_price_ret.tail(self.m6+1).pct_change(self.m6).iloc[-1]
        self.ret_res['y1_ret'] = self.index_price_ret.tail(self.y1+1).pct_change(self.y1).iloc[-1]
        self.ret_res['y3_ret'] = self.index_price_ret.tail(self.y3+1).pct_change(self.y3).iloc[-1]
        self.ret_res['y5_ret'] = self.index_price_ret.tail(self.y5+1).pct_change(self.y5).iloc[-1]
        self.ret_res['y10_ret'] = self.index_price_ret.tail(self.y10+1).pct_change(self.y10).iloc[-1]
        self.ret_res['this_s_ret'] = self._term_ret(self.index_price_ret.loc[s_date:])
        self.ret_res['this_y_ret'] = self._term_ret(self.index_price_ret.loc[y_date:])
        self.ret_res['cumulative_ret'] = self._term_ret(self.index_price_ret)
        result_ret = self._process_update_result(self.ret_res)
        result_ret = result_ret.replace(np.Inf,None).replace(-np.Inf,None).drop_duplicates(subset=['index_id','datetime'])
        return result_ret

    def _update_vol(self):
        y_date = self._get_begin_of_year(self.end_date)
        s_date = self._get_begin_of_season(self.end_date)
        self.vol_res = {}
        self.vol_res['w1_vol'] = self.index_price_vol_part.tail(self.w1).std(ddof=1)
        self.vol_res['m1_vol'] = self.index_price_vol_part.tail(self.m1).std(ddof=1)
        self.vol_res['m3_vol'] = self.index_price_vol_part.tail(self.m3).std(ddof=1)
        self.vol_res['m6_vol'] = self.index_price_vol_part.tail(self.m6).std(ddof=1)
        self.vol_res['y1_vol'] = self.index_price_vol_part.tail(self.y1).std(ddof=1)
        self.vol_res['y3_vol'] = self.index_price_vol_part.tail(self.y3).std(ddof=1)
        self.vol_res['y5_vol'] = self.index_price_vol_part.tail(self.y5).std(ddof=1)
        self.vol_res['y10_vol'] = self.index_price_vol_part.tail(self.y10).std(ddof=1)
        self.vol_res['this_s_vol'] = self._term_vol(self.index_price_vol_part.loc[s_date:])
        self.vol_res['this_y_vol'] = self._term_vol(self.index_price_vol_part.loc[y_date:])
        self.vol_res['cumulative_vol'] = self._term_vol(self.index_price_vol_part)
        result_vol = self._process_update_result(self.vol_res)
        result_vol = result_vol.replace(np.Inf,None).replace(-np.Inf,None).drop_duplicates(subset=['index_id','datetime'])
        return result_vol
        
    def process(self, end_date):
        failed_tasks = []
        try:
            #一次load 
            self.init(end_date=end_date)
            result_ret = self._update_ret()
            self._data_helper._upload_derived(result_ret, IndexReturn.__table__.name)

        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('index_return')
        
        try:
            result_vol = self._update_vol()
            self._data_helper._upload_derived(result_vol, IndexVolatility.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('index_volatility')
        return failed_tasks


from .nlp.qabot import QaBot

class WechatMessageParser(object):
    def __init__(self):
        super().__init__()
        self.bot = QaBot()

    def parse(self, msg_from, msg_to, msg_content):
        rule = self.bot.text_parser(msg_content)
        return rule